# -*-coding:utf-8 -*-
"""gklearn - utils module

Implement some methods to manage graphs
 graphfiles.py : load .gxl and .ct files
 utils.py : compute some properties on networkX graphs


"""

# info
__version__ = "0.1"
__author__ = "Benoit Gaüzère"
__date__ = "November 2017"

# from utils import graphfiles
# from utils import utils
from gklearn.utils.dataset import Dataset, split_dataset_by_target
from gklearn.utils.graph_files import load_dataset, save_dataset
from gklearn.utils.timer import Timer
from gklearn.utils.utils import get_graph_kernel_by_name
from gklearn.utils.utils import compute_gram_matrices_by_class
from gklearn.utils.utils import SpecialLabel, dummy_node, undefined_node, dummy_edge
from gklearn.utils.utils import normalize_gram_matrix, compute_distance_matrix
from gklearn.utils.utils import check_json_serializable, is_basic_python_type
from gklearn.utils.trie import Trie
from gklearn.utils.knn import knn_cv, knn_classification
from gklearn.utils.model_selection_precomputed import model_selection_for_precomputed_kernel
from gklearn.utils.iters import get_iters



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 18 16:01:24 2020

@author: ljia
"""
import numpy as np
import networkx as nx
from gklearn.ged.methods import GEDMethod
from gklearn.ged.util import LSAPESolver, misc
from gklearn.ged.env import NodeMap
	

class LSAPEBasedMethod(GEDMethod):
	
	
	def __init__(self, ged_data):
		super().__init__(ged_data)
		self._lsape_model = None # @todo: LSAPESolver::ECBP
		self._greedy_method = None # @todo: LSAPESolver::BASIC
		self._compute_lower_bound = True
		self._solve_optimally = True
		self._num_threads = 1
		self._centrality_method = 'NODE' # @todo
		self._centrality_weight = 0.7
		self._centralities = {}
		self._max_num_solutions = 1
		
		
	def populate_instance_and_run_as_util(self, g, h): #, lsape_instance):
		"""
	/*!
	 * @brief Runs the method with options specified by set_options() and provides access to constructed LSAPE instance.
	 * @param[in] g Input graph.
	 * @param[in] h Input graph.
	 * @param[out] result Result variable.
	 * @param[out] lsape_instance LSAPE instance.
	 */
		"""
		result = {'node_maps': [], 'lower_bound': 0, 'upper_bound': np.inf}
		
		# Populate the LSAPE instance and set up the solver.
		nb1, nb2 = nx.number_of_nodes(g), nx.number_of_nodes(h)
		lsape_instance = np.ones((nb1 + nb2, nb1 + nb2)) * np.inf
# 		lsape_instance = np.empty((nx.number_of_nodes(g) + 1, nx.number_of_nodes(h) + 1))
		self.populate_instance(g, h, lsape_instance)
		
# 		nb1, nb2 = nx.number_of_nodes(g), nx.number_of_nodes(h)
# 		lsape_instance_new = np.empty((nb1 + nb2, nb1 + nb2)) * np.inf
# 		lsape_instance_new[nb1:, nb2:] = 0
# 		lsape_instance_new[0:nb1, 0:nb2] = lsape_instance[0:nb1, 0:nb2]
# 		for i in range(nb1): # all u's neighbor
# 			lsape_instance_new[i, nb2 + i] = lsape_instance[i, nb2]
# 		for i in range(nb2): # all u's neighbor
# 			lsape_instance_new[nb1 + i, i] = lsape_instance[nb2, i]
# 		lsape_solver = LSAPESolver(lsape_instance_new)
		
		lsape_solver = LSAPESolver(lsape_instance)
		
		# Solve the LSAPE instance.
		if self._solve_optimally:
			lsape_solver.set_model(self._lsape_model)
		else:
			lsape_solver.set_greedy_method(self._greedy_method)
		lsape_solver.solve(self._max_num_solutions)
		
		# Compute and store lower and upper bound.
		if self._compute_lower_bound and self._solve_optimally:
			result['lower_bound'] = lsape_solver.minimal_cost() * self._lsape_lower_bound_scaling_factor(g, h) # @todo: test
		
		for solution_id in range(0, lsape_solver.num_solutions()):
			result['node_maps'].append(NodeMap(nx.number_of_nodes(g), nx.number_of_nodes(h)))
			misc.construct_node_map_from_solver(lsape_solver, result['node_maps'][-1], solution_id)
			self._ged_data.compute_induced_cost(g, h, result['node_maps'][-1])
			
		# Add centralities and reoptimize.
		if self._centrality_weight > 0 and self._centrality_method != 'NODE':
			print('This is not implemented.')
			pass # @todo
			
		# Sort the node maps and set the upper bound.
		if len(result['node_maps']) > 1 or len(result['node_maps']) > self._max_num_solutions:
			print('This is not implemented.') # @todo:
			pass
		if len(result['node_maps']) == 0:
			result['upper_bound'] = np.inf
		else:
			result['upper_bound'] = result['node_maps'][0].induced_cost()
					
		return result
			
		
	
	def populate_instance(self, g, h, lsape_instance):
		"""
	/*!
	 * @brief Populates the LSAPE instance.
	 * @param[in] g Input graph.
	 * @param[in] h Input graph.
	 * @param[out] lsape_instance LSAPE instance.
	 */
		"""
		if not self._initialized:
			pass
		# @todo: if (not this->initialized_) {
		self._lsape_populate_instance(g, h, lsape_instance)
		lsape_instance[nx.number_of_nodes(g):, nx.number_of_nodes(h):] = 0
# 		lsape_instance[nx.number_of_nodes(g), nx.number_of_nodes(h)] = 0
		
	
	###########################################################################
	# Member functions inherited from GEDMethod.
	###########################################################################
	
	
	def _ged_init(self):
		self._lsape_pre_graph_init(False)
		for graph in self._ged_data._graphs:
			self._init_graph(graph)
		self._lsape_init()
		
		
	def _ged_run(self, g, h):
# 		lsape_instance = np.empty((0, 0))
		result = self.populate_instance_and_run_as_util(g, h) # , lsape_instance)
		return result
		
		
	def _ged_parse_option(self, option, arg):
		is_valid_option = False
		
		if option == 'threads': # @todo: try.. catch...
			self._num_threads = arg
			is_valid_option = True
		elif option == 'lsape_model':
			self._lsape_model = arg # @todo
			is_valid_option = True
		elif option == 'greedy_method':
			self._greedy_method = arg # @todo
			is_valid_option = True
		elif option == 'optimal':
			self._solve_optimally = arg # @todo
			is_valid_option = True
		elif option == 'centrality_method':
			self._centrality_method = arg # @todo
			is_valid_option = True
		elif option == 'centrality_weight':
			self._centrality_weight = arg # @todo
			is_valid_option = True
		elif option == 'max_num_solutions':
			if arg == 'ALL':
				self._max_num_solutions = -1
			else:				
				self._max_num_solutions = arg # @todo
			is_valid_option = True
			
		is_valid_option = is_valid_option or self._lsape_parse_option(option, arg)
		is_valid_option = True # @todo: this is not in the C++ code.
		return is_valid_option
		
		
	def _ged_set_default_options(self):
		self._lsape_model = None # @todo: LSAPESolver::ECBP
		self._greedy_method = None # @todo: LSAPESolver::BASIC
		self._solve_optimally = True
		self._num_threads = 1
		self._centrality_method = 'NODE' # @todo
		self._centrality_weight = 0.7
		self._max_num_solutions = 1
		
		
	###########################################################################
	# Private helper member functions.
	###########################################################################
	
	
	def _init_graph(self, graph):
		if self._centrality_method != 'NODE':
			self._init_centralities(graph) # @todo
		self._lsape_init_graph(graph)
	
	
	###########################################################################
	# Virtual member functions to be overridden by derived classes.
	###########################################################################
	
	
	def _lsape_init(self):
		"""
	/*!
	 * @brief Initializes the method after initializing the global variables for the graphs.
	 * @note Must be overridden by derived classes of ged::LSAPEBasedMethod that require custom initialization.
	 */
		"""
		pass
		
		
	def _lsape_parse_option(self, option, arg):
		"""
	/*!
	 * @brief Parses one option that is not among the ones shared by all derived classes of ged::LSAPEBasedMethod.
	 * @param[in] option The name of the option.
	 * @param[in] arg The argument of the option.
	 * @return Returns true if @p option is a valid option name for the method and false otherwise.
	 * @note Must be overridden by derived classes of ged::LSAPEBasedMethod that have options that are not among the ones shared by all derived classes of ged::LSAPEBasedMethod.
	 */
		"""
		return False
		
		
	def _lsape_set_default_options(self):
		"""
	/*!
	 * @brief Sets all options that are not among the ones shared by all derived classes of ged::LSAPEBasedMethod to default values.
	 * @note Must be overridden by derived classes of ged::LSAPEBasedMethod that have options that are not among the ones shared by all derived classes of ged::LSAPEBasedMethod.
	 */
		"""
		pass
	
	
	def _lsape_populate_instance(self, g, h, lsape_instance):
		"""
	/*!
	 * @brief Populates the LSAPE instance.
	 * @param[in] g Input graph.
	 * @param[in] h Input graph.
	 * @param[out] lsape_instance LSAPE instance of size (n + 1) x (m + 1), where n and m are the number of nodes in @p g and @p h. The last row and the last column represent insertion and deletion.
	 * @note Must be overridden by derived classes of ged::LSAPEBasedMethod.
	 */
		"""
		pass
	
	
	def _lsape_init_graph(self, graph):
		"""
	/*!
	 * @brief Initializes global variables for one graph.
	 * @param[in] graph Graph for which the global variables have to be initialized.
	 * @note Must be overridden by derived classes of ged::LSAPEBasedMethod that require to initialize custom global variables.
	 */
		"""
		pass
	
	
	def _lsape_pre_graph_init(self, called_at_runtime):
		"""
	/*!
	 * @brief Initializes the method at runtime or during initialization before initializing the global variables for the graphs.
	 * @param[in] called_at_runtime Equals @p true if called at runtime and @p false if called during initialization.
	 * @brief Must be overridden by derived classes of ged::LSAPEBasedMethod that require default initialization at runtime before initializing the global variables for the graphs.
	 */
		"""
		pass


"""Tests of graph kernels.
"""

import multiprocessing
import numpy as np


##############################################################################

def chooseDataset(ds_name):
	"""Choose dataset according to name.
	"""
	from gklearn.dataset import Dataset
	import os

	current_path = os.path.dirname(os.path.realpath(__file__)) + '/'
	root = current_path + '../../datasets/'

	# no labels at all.
	if ds_name == 'Alkane_unlabeled':
		dataset = Dataset('Alkane_unlabeled', root=root)
		dataset.trim_dataset(edge_required=False)
		dataset.cut_graphs(range(1, 10))
	# node symbolic labels only.
	elif ds_name == 'Acyclic':
		dataset = Dataset('Acyclic', root=root)
		dataset.trim_dataset(edge_required=False)
	# node non-symbolic labels only.
	elif ds_name == 'Letter-med':
		dataset = Dataset('Letter-med', root=root)
		dataset.trim_dataset(edge_required=False)
	# node symbolic + non-symbolic labels + edge symbolic labels.
	elif ds_name == 'AIDS':
		dataset = Dataset('AIDS', root=root)
		dataset.trim_dataset(edge_required=False)
	# node non-symbolic labels + edge non-symbolic labels.
	elif ds_name == 'Fingerprint':
		dataset = Dataset('Fingerprint', root=root)
		dataset.trim_dataset(edge_required=True)
	# edge symbolic only.
	elif ds_name == 'MAO':
		dataset = Dataset('MAO', root=root)
		dataset.trim_dataset(edge_required=True)
		irrelevant_labels = {'node_labels': ['atom_symbol'], 'node_attrs': ['x', 'y']}
		dataset.remove_labels(**irrelevant_labels)
	# edge non-symbolic labels only.
	elif ds_name == 'Fingerprint_edge':
		dataset = Dataset('Fingerprint', root=root)
		dataset.trim_dataset(edge_required=True)
		irrelevant_labels = {'edge_attrs': ['orient', 'angle']}
		dataset.remove_labels(**irrelevant_labels)
	# node symbolic and non-symbolic labels + edge symbolic and non-symbolic labels.
	elif ds_name == 'Cuneiform':
		dataset = Dataset('Cuneiform', root=root)
		dataset.trim_dataset(edge_required=True)

	dataset.cut_graphs(range(0, 3))

	return dataset


def assert_equality(compute_fun, **kwargs):
	"""Check if outputs are the same using different methods to compute.

	Parameters
	----------
	compute_fun : function
		The function to compute the kernel, with the same key word arguments as
		kwargs.
	**kwargs : dict
		The key word arguments over the grid of which the kernel results are
		compared.

	Returns
	-------
	None.
	"""
	from sklearn.model_selection import ParameterGrid
	param_grid = ParameterGrid(kwargs)

	result_lists = [[], [], []]
	for params in list(param_grid):
		results = compute_fun(**params)
		for rs, lst in zip(results, result_lists):
			lst.append(rs)
	for lst in result_lists:
		for i in range(len(lst[:-1])):
			assert np.array_equal(lst[i], lst[i + 1])


# @pytest.mark.parametrize('parallel', ['imap_unordered', None])
def my_test_CommonWalk(ds_name, weight, compute_method):
	"""Test common walk kernel.
	"""
	def compute(parallel=None):
		from gklearn.kernels import CommonWalk
		import networkx as nx

		dataset = chooseDataset(ds_name)
		dataset.load_graphs([g for g in dataset.graphs if nx.number_of_nodes(g) > 1])

		try:
			graph_kernel = CommonWalk(node_labels=dataset.node_labels,
						edge_labels=dataset.edge_labels,
						ds_infos=dataset.get_dataset_infos(keys=['directed']),
						weight=weight,
						compute_method=compute_method)
			gram_matrix, run_time = graph_kernel.compute(dataset.graphs,
				parallel=parallel, n_jobs=multiprocessing.cpu_count(), verbose=True)
			kernel_list, run_time = graph_kernel.compute(dataset.graphs[0], dataset.graphs[1:],
				parallel=parallel, n_jobs=multiprocessing.cpu_count(), verbose=True)
			kernel, run_time = graph_kernel.compute(dataset.graphs[0], dataset.graphs[1],
				parallel=parallel, n_jobs=multiprocessing.cpu_count(), verbose=True)

		except Exception as exception:
			print(repr(exception))
			assert False, exception
		else:
			return gram_matrix, kernel_list, kernel

	assert_equality(compute, parallel=[None, 'imap_unordered'])


if __name__ == "__main__":
 	# test_list_graph_kernels()
#	test_spkernel('Alkane_unlabeled', 'imap_unordered')
 	# test_ShortestPath('Alkane_unlabeled')
# 	test_StructuralSP('Fingerprint_edge', 'imap_unordered')
 	# test_StructuralSP('Acyclic')
# 	test_StructuralSP('Cuneiform', None)
#  	test_WLSubtree('MAO') # 'Alkane_unlabeled', 'Acyclic', 'AIDS'
#	test_RandomWalk('Acyclic', 'sylvester', None, 'imap_unordered')
#	test_RandomWalk('Acyclic', 'conjugate', None, 'imap_unordered')
#	test_RandomWalk('Acyclic', 'fp', None, None)
#	test_RandomWalk('Acyclic', 'spectral', 'exp', 'imap_unordered')
 	# test_CommonWalk('Acyclic', 0.01, 'geo')
	my_test_CommonWalk('AIDS', 0.01, 'geo')
    # test_Marginalized('Acyclic', False)
 	# test_ShortestPath('Acyclic')
# 	 test_PathUpToH('Acyclic', 'MinMax')
 	# test_Treelet('AIDS')
# 	test_SylvesterEquation('Acyclic')
# 	test_ConjugateGradient('Acyclic')
# 	test_FixedPoint('Acyclic')
# 	test_SpectralDecomposition('Acyclic', 'exp')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Mar 26 18:48:27 2020

@author: ljia
"""
import numpy as np
import networkx as nx
from gklearn.utils.graph_files import load_dataset
import os


class Dataset(object):


	def __init__(self, filename=None, filename_targets=None, **kwargs):
		import warnings
		warnings.simplefilter('always', DeprecationWarning)
		warnings.warn('This class has been moved to "gklearn.dataset" module. The class "gklearn.utils.dataset.Dataset" has not been maintained since Nov 12th, 2020 (version 0.2.1) and will be removed since version 0.4.0.', DeprecationWarning)

		if filename is None:
			self._graphs = None
			self._targets = None
			self._node_labels = None
			self._edge_labels = None
			self._node_attrs = None
			self._edge_attrs = None
		else:
			self.load_dataset(filename, filename_targets=filename_targets, **kwargs)

		self._substructures = None
		self._node_label_dim = None
		self._edge_label_dim = None
		self._directed = None
		self._dataset_size = None
		self._total_node_num = None
		self._ave_node_num = None
		self._min_node_num = None
		self._max_node_num = None
		self._total_edge_num = None
		self._ave_edge_num = None
		self._min_edge_num = None
		self._max_edge_num = None
		self._ave_node_degree = None
		self._min_node_degree = None
		self._max_node_degree = None
		self._ave_fill_factor = None
		self._min_fill_factor = None
		self._max_fill_factor = None
		self._node_label_nums = None
		self._edge_label_nums = None
		self._node_attr_dim = None
		self._edge_attr_dim = None
		self._class_number = None


	def load_dataset(self, filename, filename_targets=None, **kwargs):
		self._graphs, self._targets, label_names = load_dataset(filename, filename_targets=filename_targets, **kwargs)
		self._node_labels = label_names['node_labels']
		self._node_attrs = label_names['node_attrs']
		self._edge_labels = label_names['edge_labels']
		self._edge_attrs = label_names['edge_attrs']
		self.clean_labels()


	def load_graphs(self, graphs, targets=None):
		# this has to be followed by set_labels().
		self._graphs = graphs
		self._targets = targets
#		self.set_labels_attrs() # @todo


	def load_predefined_dataset(self, ds_name):
		current_path = os.path.dirname(os.path.realpath(__file__)) + '/'
		if ds_name == 'Acyclic':
			ds_file = current_path + '../../datasets/Acyclic/dataset_bps.ds'
			self._graphs, self._targets, label_names = load_dataset(ds_file)
		elif ds_name == 'AIDS':
			ds_file = current_path + '../../datasets/AIDS/AIDS_A.txt'
			self._graphs, self._targets, label_names = load_dataset(ds_file)
		elif ds_name == 'Alkane':
			ds_file = current_path + '../../datasets/Alkane/dataset.ds'
			fn_targets = current_path + '../../datasets/Alkane/dataset_boiling_point_names.txt'
			self._graphs, self._targets, label_names = load_dataset(ds_file, filename_targets=fn_targets)
		elif ds_name == 'COIL-DEL':
			ds_file = current_path + '../../datasets/COIL-DEL/COIL-DEL_A.txt'
			self._graphs, self._targets, label_names = load_dataset(ds_file)
		elif ds_name == 'COIL-RAG':
			ds_file = current_path + '../../datasets/COIL-RAG/COIL-RAG_A.txt'
			self._graphs, self._targets, label_names = load_dataset(ds_file)
		elif ds_name == 'COLORS-3':
			ds_file = current_path + '../../datasets/COLORS-3/COLORS-3_A.txt'
			self._graphs, self._targets, label_names = load_dataset(ds_file)
		elif ds_name == 'Cuneiform':
			ds_file = current_path + '../../datasets/Cuneiform/Cuneiform_A.txt'
			self._graphs, self._targets, label_names = load_dataset(ds_file)
		elif ds_name == 'DD':
			ds_file = current_path + '../../datasets/DD/DD_A.txt'
			self._graphs, self._targets, label_names = load_dataset(ds_file)
		elif ds_name == 'ENZYMES':
			ds_file = current_path + '../../datasets/ENZYMES_txt/ENZYMES_A_sparse.txt'
			self._graphs, self._targets, label_names = load_dataset(ds_file)
		elif ds_name == 'Fingerprint':
			ds_file = current_path + '../../datasets/Fingerprint/Fingerprint_A.txt'
			self._graphs, self._targets, label_names = load_dataset(ds_file)
		elif ds_name == 'FRANKENSTEIN':
			ds_file = current_path + '../../datasets/FRANKENSTEIN/FRANKENSTEIN_A.txt'
			self._graphs, self._targets, label_names = load_dataset(ds_file)
		elif ds_name == 'Letter-high': # node non-symb
			ds_file = current_path + '../../datasets/Letter-high/Letter-high_A.txt'
			self._graphs, self._targets, label_names = load_dataset(ds_file)
		elif ds_name == 'Letter-low': # node non-symb
			ds_file = current_path + '../../datasets/Letter-low/Letter-low_A.txt'
			self._graphs, self._targets, label_names = load_dataset(ds_file)
		elif ds_name == 'Letter-med': # node non-symb
			ds_file = current_path + '../../datasets/Letter-med/Letter-med_A.txt'
			self._graphs, self._targets, label_names = load_dataset(ds_file)
		elif ds_name == 'MAO':
			ds_file = current_path + '../../datasets/MAO/dataset.ds'
			self._graphs, self._targets, label_names = load_dataset(ds_file)
		elif ds_name == 'Monoterpenoides':
			ds_file = current_path + '../../datasets/Monoterpenoides/dataset_10+.ds'
			self._graphs, self._targets, label_names = load_dataset(ds_file)
		elif ds_name == 'MUTAG':
			ds_file = current_path + '../../datasets/MUTAG/MUTAG_A.txt'
			self._graphs, self._targets, label_names = load_dataset(ds_file)
		elif ds_name == 'NCI1':
			ds_file = current_path + '../../datasets/NCI1/NCI1_A.txt'
			self._graphs, self._targets, label_names = load_dataset(ds_file)
		elif ds_name == 'NCI109':
			ds_file = current_path + '../../datasets/NCI109/NCI109_A.txt'
			self._graphs, self._targets, label_names = load_dataset(ds_file)
		elif ds_name == 'PAH':
			ds_file = current_path + '../../datasets/PAH/dataset.ds'
			self._graphs, self._targets, label_names = load_dataset(ds_file)
		elif ds_name == 'SYNTHETIC':
			pass
		elif ds_name == 'SYNTHETICnew':
			ds_file = current_path + '../../datasets/SYNTHETICnew/SYNTHETICnew_A.txt'
			self._graphs, self._targets, label_names = load_dataset(ds_file)
		elif ds_name == 'Synthie':
			pass
		else:
			raise Exception('The dataset name "', ds_name, '" is not pre-defined.')

		self._node_labels = label_names['node_labels']
		self._node_attrs = label_names['node_attrs']
		self._edge_labels = label_names['edge_labels']
		self._edge_attrs = label_names['edge_attrs']
		self.clean_labels()


	def set_labels(self, node_labels=[], node_attrs=[], edge_labels=[], edge_attrs=[]):
		self._node_labels = node_labels
		self._node_attrs = node_attrs
		self._edge_labels = edge_labels
		self._edge_attrs = edge_attrs


	def set_labels_attrs(self, node_labels=None, node_attrs=None, edge_labels=None, edge_attrs=None):
		# @todo: remove labels which have only one possible values.
		if node_labels is None:
			self._node_labels = self._graphs[0].graph['node_labels']
#			# graphs are considered node unlabeled if all nodes have the same label.
#			infos.update({'node_labeled': is_nl if node_label_num > 1 else False})
		if node_attrs is None:
			self._node_attrs = self._graphs[0].graph['node_attrs']
#		for G in Gn:
#			for n in G.nodes(data=True):
#				if 'attributes' in n[1]:
#					return len(n[1]['attributes'])
#		return 0
		if edge_labels is None:
			self._edge_labels = self._graphs[0].graph['edge_labels']
#			# graphs are considered edge unlabeled if all edges have the same label.
#			infos.update({'edge_labeled': is_el if edge_label_num > 1 else False})
		if edge_attrs is None:
			self._edge_attrs = self._graphs[0].graph['edge_attrs']
#		for G in Gn:
#			if nx.number_of_edges(G) > 0:
#				for e in G.edges(data=True):
#					if 'attributes' in e[2]:
#						return len(e[2]['attributes'])
#		return 0


	def get_dataset_infos(self, keys=None, params=None):
		"""Computes and returns the structure and property information of the graph dataset.

		Parameters
		----------
		keys : list, optional
			A list of strings which indicate which informations will be returned. The
			possible choices includes:

			'substructures': sub-structures graphs contains, including 'linear', 'non
		linear' and 'cyclic'.

			'node_label_dim': whether vertices have symbolic labels.

			'edge_label_dim': whether egdes have symbolic labels.

			'directed': whether graphs in dataset are directed.

			'dataset_size': number of graphs in dataset.

			'total_node_num': total number of vertices of all graphs in dataset.

			'ave_node_num': average number of vertices of graphs in dataset.

			'min_node_num': minimum number of vertices of graphs in dataset.

			'max_node_num': maximum number of vertices of graphs in dataset.

			'total_edge_num': total number of edges of all graphs in dataset.

			'ave_edge_num': average number of edges of graphs in dataset.

			'min_edge_num': minimum number of edges of graphs in dataset.

			'max_edge_num': maximum number of edges of graphs in dataset.

			'ave_node_degree': average vertex degree of graphs in dataset.

			'min_node_degree': minimum vertex degree of graphs in dataset.

			'max_node_degree': maximum vertex degree of graphs in dataset.

			'ave_fill_factor': average fill factor (number_of_edges /
		(number_of_nodes ** 2)) of graphs in dataset.

			'min_fill_factor': minimum fill factor of graphs in dataset.

			'max_fill_factor': maximum fill factor of graphs in dataset.

			'node_label_nums': list of numbers of symbolic vertex labels of graphs in dataset.

			'edge_label_nums': list number of symbolic edge labels of graphs in dataset.

			'node_attr_dim': number of dimensions of non-symbolic vertex labels.
		Extracted from the 'attributes' attribute of graph nodes.

			'edge_attr_dim': number of dimensions of non-symbolic edge labels.
		Extracted from the 'attributes' attribute of graph edges.

			'class_number': number of classes. Only available for classification problems.

			'all_degree_entropy': the entropy of degree distribution of each graph.

			'ave_degree_entropy': the average entropy of degree distribution of all graphs.

			All informations above will be returned if `keys` is not given.

		params: dict of dict, optional
			A dictinary which contains extra parameters for each possible
			element in ``keys``.

		Return
		------
		dict
			Information of the graph dataset keyed by `keys`.
		"""
		infos = {}

		if keys == None:
			keys = [
				'substructures',
				'node_label_dim',
				'edge_label_dim',
				'directed',
				'dataset_size',
				'total_node_num',
				'ave_node_num',
				'min_node_num',
				'max_node_num',
				'total_edge_num',
				'ave_edge_num',
				'min_edge_num',
				'max_edge_num',
				'ave_node_degree',
				'min_node_degree',
				'max_node_degree',
				'ave_fill_factor',
				'min_fill_factor',
				'max_fill_factor',
				'node_label_nums',
				'edge_label_nums',
				'node_attr_dim',
				'edge_attr_dim',
				'class_number',
				'all_degree_entropy',
				'ave_degree_entropy'
			]

		# dataset size
		if 'dataset_size' in keys:
			if self._dataset_size is None:
				self._dataset_size = self._get_dataset_size()
			infos['dataset_size'] = self._dataset_size

		# graph node number
		if any(i in keys for i in ['total_node_num', 'ave_node_num', 'min_node_num', 'max_node_num']):
			all_node_nums = self._get_all_node_nums()

		if 'total_node_num' in keys:
			if self._total_node_num is None:
				self._total_node_num = self._get_total_node_num(all_node_nums)
			infos['total_node_num'] = self._total_node_num

		if 'ave_node_num' in keys:
			if self._ave_node_num is None:
				self._ave_node_num = self._get_ave_node_num(all_node_nums)
			infos['ave_node_num'] = self._ave_node_num

		if 'min_node_num' in keys:
			if self._min_node_num is None:
				self._min_node_num = self._get_min_node_num(all_node_nums)
			infos['min_node_num'] = self._min_node_num

		if 'max_node_num' in keys:
			if self._max_node_num is None:
				self._max_node_num = self._get_max_node_num(all_node_nums)
			infos['max_node_num'] = self._max_node_num

		# graph edge number
		if any(i in keys for i in ['total_edge_num', 'ave_edge_num', 'min_edge_num', 'max_edge_num']):
			all_edge_nums = self._get_all_edge_nums()

		if 'total_edge_num' in keys:
			if self._total_edge_num is None:
				self._total_edge_num = self._get_total_edge_num(all_edge_nums)
			infos['total_edge_num'] = self._total_edge_num

		if 'ave_edge_num' in keys:
			if self._ave_edge_num is None:
				self._ave_edge_num = self._get_ave_edge_num(all_edge_nums)
			infos['ave_edge_num'] = self._ave_edge_num

		if 'max_edge_num' in keys:
			if self._max_edge_num is None:
				self._max_edge_num = self._get_max_edge_num(all_edge_nums)
			infos['max_edge_num'] = self._max_edge_num

		if 'min_edge_num' in keys:
			if self._min_edge_num is None:
				self._min_edge_num = self._get_min_edge_num(all_edge_nums)
			infos['min_edge_num'] = self._min_edge_num

		# label number
		if 'node_label_dim' in keys:
			if self._node_label_dim is None:
				self._node_label_dim = self._get_node_label_dim()
			infos['node_label_dim'] = self._node_label_dim

		if 'node_label_nums' in keys:
			if self._node_label_nums is None:
				self._node_label_nums = {}
				for node_label in self._node_labels:
					self._node_label_nums[node_label] = self._get_node_label_num(node_label)
			infos['node_label_nums'] = self._node_label_nums

		if 'edge_label_dim' in keys:
			if self._edge_label_dim is None:
				self._edge_label_dim = self._get_edge_label_dim()
			infos['edge_label_dim'] = self._edge_label_dim

		if 'edge_label_nums' in keys:
			if self._edge_label_nums is None:
				self._edge_label_nums = {}
				for edge_label in self._edge_labels:
					self._edge_label_nums[edge_label] = self._get_edge_label_num(edge_label)
			infos['edge_label_nums'] = self._edge_label_nums

		if 'directed' in keys or 'substructures' in keys:
			if self._directed is None:
				self._directed = self._is_directed()
			infos['directed'] = self._directed

		# node degree
		if any(i in keys for i in ['ave_node_degree', 'max_node_degree', 'min_node_degree']):
			all_node_degrees = self._get_all_node_degrees()

		if 'ave_node_degree' in keys:
			if self._ave_node_degree is None:
				self._ave_node_degree = self._get_ave_node_degree(all_node_degrees)
			infos['ave_node_degree'] = self._ave_node_degree

		if 'max_node_degree' in keys:
			if self._max_node_degree is None:
				self._max_node_degree = self._get_max_node_degree(all_node_degrees)
			infos['max_node_degree'] = self._max_node_degree

		if 'min_node_degree' in keys:
			if self._min_node_degree is None:
				self._min_node_degree = self._get_min_node_degree(all_node_degrees)
			infos['min_node_degree'] = self._min_node_degree

		# fill factor
		if any(i in keys for i in ['ave_fill_factor', 'max_fill_factor', 'min_fill_factor']):
			all_fill_factors = self._get_all_fill_factors()

		if 'ave_fill_factor' in keys:
			if self._ave_fill_factor is None:
				self._ave_fill_factor = self._get_ave_fill_factor(all_fill_factors)
			infos['ave_fill_factor'] = self._ave_fill_factor

		if 'max_fill_factor' in keys:
			if self._max_fill_factor is None:
				self._max_fill_factor = self._get_max_fill_factor(all_fill_factors)
			infos['max_fill_factor'] = self._max_fill_factor

		if 'min_fill_factor' in keys:
			if self._min_fill_factor is None:
				self._min_fill_factor = self._get_min_fill_factor(all_fill_factors)
			infos['min_fill_factor'] = self._min_fill_factor

		if 'substructures' in keys:
			if self._substructures is None:
				self._substructures = self._get_substructures()
			infos['substructures'] = self._substructures

		if 'class_number' in keys:
			if self._class_number is None:
				self._class_number = self._get_class_number()
			infos['class_number'] = self._class_number

		if 'node_attr_dim' in keys:
			if self._node_attr_dim is None:
				self._node_attr_dim = self._get_node_attr_dim()
			infos['node_attr_dim'] = self._node_attr_dim

		if 'edge_attr_dim' in keys:
			if self._edge_attr_dim is None:
				self._edge_attr_dim = self._get_edge_attr_dim()
			infos['edge_attr_dim'] = self._edge_attr_dim

		# entropy of degree distribution.

		if 'all_degree_entropy' in keys:
			if params is not None and ('all_degree_entropy' in params) and ('base' in params['all_degree_entropy']):
				base = params['all_degree_entropy']['base']
			else:
				base = None
			infos['all_degree_entropy'] = self._compute_all_degree_entropy(base=base)

		if 'ave_degree_entropy' in keys:
			if params is not None and ('ave_degree_entropy' in params) and ('base' in params['ave_degree_entropy']):
				base = params['ave_degree_entropy']['base']
			else:
				base = None
			infos['ave_degree_entropy'] = np.mean(self._compute_all_degree_entropy(base=base))

		return infos


	def print_graph_infos(self, infos):
		from collections import OrderedDict
		keys = list(infos.keys())
		print(OrderedDict(sorted(infos.items(), key=lambda i: keys.index(i[0]))))


	def remove_labels(self, node_labels=[], edge_labels=[], node_attrs=[], edge_attrs=[]):
		node_labels = [item for item in node_labels if item in self._node_labels]
		edge_labels = [item for item in edge_labels if item in self._edge_labels]
		node_attrs = [item for item in node_attrs if item in self._node_attrs]
		edge_attrs = [item for item in edge_attrs if item in self._edge_attrs]

		for g in self._graphs:
			for nd in g.nodes():
				for nl in node_labels:
					del g.nodes[nd][nl]
				for na in node_attrs:
					del g.nodes[nd][na]
			for ed in g.edges():
				for el in edge_labels:
					del g.edges[ed][el]
				for ea in edge_attrs:
					del g.edges[ed][ea]
		if len(node_labels) > 0:
			self._node_labels = [nl for nl in self._node_labels if nl not in node_labels]
		if len(edge_labels) > 0:
			self._edge_labels = [el for el in self._edge_labels if el not in edge_labels]
		if len(node_attrs) > 0:
			self._node_attrs = [na for na in self._node_attrs if na not in node_attrs]
		if len(edge_attrs) > 0:
			self._edge_attrs = [ea for ea in self._edge_attrs if ea not in edge_attrs]


	def clean_labels(self):
		labels = []
		for name in self._node_labels:
			label = set()
			for G in self._graphs:
				label = label | set(nx.get_node_attributes(G, name).values())
				if len(label) > 1:
					labels.append(name)
					break
			if len(label) < 2:
				for G in self._graphs:
					for nd in G.nodes():
						del G.nodes[nd][name]
		self._node_labels = labels

		labels = []
		for name in self._edge_labels:
			label = set()
			for G in self._graphs:
				label = label | set(nx.get_edge_attributes(G, name).values())
				if len(label) > 1:
					labels.append(name)
					break
			if len(label) < 2:
				for G in self._graphs:
					for ed in G.edges():
						del G.edges[ed][name]
		self._edge_labels = labels

		labels = []
		for name in self._node_attrs:
			label = set()
			for G in self._graphs:
				label = label | set(nx.get_node_attributes(G, name).values())
				if len(label) > 1:
					labels.append(name)
					break
			if len(label) < 2:
				for G in self._graphs:
					for nd in G.nodes():
						del G.nodes[nd][name]
		self._node_attrs = labels

		labels = []
		for name in self._edge_attrs:
			label = set()
			for G in self._graphs:
				label = label | set(nx.get_edge_attributes(G, name).values())
				if len(label) > 1:
					labels.append(name)
					break
			if len(label) < 2:
				for G in self._graphs:
					for ed in G.edges():
						del G.edges[ed][name]
		self._edge_attrs = labels


	def cut_graphs(self, range_):
		self._graphs = [self._graphs[i] for i in range_]
		if self._targets is not None:
			self._targets = [self._targets[i] for i in range_]
		self.clean_labels()


	def trim_dataset(self, edge_required=False):
		if edge_required:
			trimed_pairs = [(idx, g) for idx, g in enumerate(self._graphs) if (nx.number_of_nodes(g) != 0 and nx.number_of_edges(g) != 0)]
		else:
			trimed_pairs = [(idx, g) for idx, g in enumerate(self._graphs) if nx.number_of_nodes(g) != 0]
		idx = [p[0] for p in trimed_pairs]
		self._graphs = [p[1] for p in trimed_pairs]
		self._targets = [self._targets[i] for i in idx]
		self.clean_labels()


	def copy(self):
		dataset = Dataset()
		graphs = [g.copy() for g in self._graphs] if self._graphs is not None else None
		target = self._targets.copy() if self._targets is not None else None
		node_labels = self._node_labels.copy() if self._node_labels is not None else None
		node_attrs = self._node_attrs.copy() if self._node_attrs is not None else None
		edge_labels = self._edge_labels.copy() if self._edge_labels is not None else None
		edge_attrs = self._edge_attrs.copy() if self._edge_attrs is not None else None
		dataset.load_graphs(graphs, target)
		dataset.set_labels(node_labels=node_labels, node_attrs=node_attrs, edge_labels=edge_labels, edge_attrs=edge_attrs)
		# @todo: clean_labels and add other class members?
		return dataset


	def get_all_node_labels(self):
		node_labels = []
		for g in self._graphs:
			for n in g.nodes():
				nl = tuple(g.nodes[n].items())
				if nl not in node_labels:
					node_labels.append(nl)
		return node_labels


	def get_all_edge_labels(self):
		edge_labels = []
		for g in self._graphs:
			for e in g.edges():
				el = tuple(g.edges[e].items())
				if el not in edge_labels:
					edge_labels.append(el)
		return edge_labels


	def _get_dataset_size(self):
		return len(self._graphs)


	def _get_all_node_nums(self):
		return [nx.number_of_nodes(G) for G in self._graphs]


	def _get_total_node_nums(self, all_node_nums):
		return np.sum(all_node_nums)


	def _get_ave_node_num(self, all_node_nums):
		return np.mean(all_node_nums)


	def _get_min_node_num(self, all_node_nums):
		return np.amin(all_node_nums)


	def _get_max_node_num(self, all_node_nums):
		return np.amax(all_node_nums)


	def _get_all_edge_nums(self):
		return [nx.number_of_edges(G) for G in self._graphs]


	def _get_total_edge_nums(self, all_edge_nums):
		return np.sum(all_edge_nums)


	def _get_ave_edge_num(self, all_edge_nums):
		return np.mean(all_edge_nums)


	def _get_min_edge_num(self, all_edge_nums):
		return np.amin(all_edge_nums)


	def _get_max_edge_num(self, all_edge_nums):
		return np.amax(all_edge_nums)


	def _get_node_label_dim(self):
		return len(self._node_labels)


	def _get_node_label_num(self, node_label):
		nl = set()
		for G in self._graphs:
			nl = nl | set(nx.get_node_attributes(G, node_label).values())
		return len(nl)


	def _get_edge_label_dim(self):
		return len(self._edge_labels)


	def _get_edge_label_num(self, edge_label):
		el = set()
		for G in self._graphs:
			el = el | set(nx.get_edge_attributes(G, edge_label).values())
		return len(el)


	def _is_directed(self):
		return nx.is_directed(self._graphs[0])


	def _get_all_node_degrees(self):
		return [np.mean(list(dict(G.degree()).values())) for G in self._graphs]


	def _get_ave_node_degree(self, all_node_degrees):
		return np.mean(all_node_degrees)


	def _get_max_node_degree(self, all_node_degrees):
		return np.amax(all_node_degrees)


	def _get_min_node_degree(self, all_node_degrees):
		return np.amin(all_node_degrees)


	def _get_all_fill_factors(self):
		"""Get fill factor, the number of non-zero entries in the adjacency matrix.

		Returns
		-------
		list[float]
			List of fill factors for all graphs.
		"""
		return [nx.number_of_edges(G) / (nx.number_of_nodes(G) ** 2) for G in self._graphs]


	def _get_ave_fill_factor(self, all_fill_factors):
		return np.mean(all_fill_factors)


	def _get_max_fill_factor(self, all_fill_factors):
		return np.amax(all_fill_factors)


	def _get_min_fill_factor(self, all_fill_factors):
		return np.amin(all_fill_factors)


	def _get_substructures(self):
		subs = set()
		for G in self._graphs:
			degrees = list(dict(G.degree()).values())
			if any(i == 2 for i in degrees):
				subs.add('linear')
			if np.amax(degrees) >= 3:
				subs.add('non linear')
			if 'linear' in subs and 'non linear' in subs:
				break

		if self._directed:
			for G in self._graphs:
				if len(list(nx.find_cycle(G))) > 0:
					subs.add('cyclic')
					break
			# else:
			#	 # @todo: this method does not work for big graph with large amount of edges like D&D, try a better way.
			#	 upper = np.amin([nx.number_of_edges(G) for G in Gn]) * 2 + 10
			#	 for G in Gn:
			#		 if (nx.number_of_edges(G) < upper):
			#			 cyc = list(nx.simple_cycles(G.to_directed()))
			#			 if any(len(i) > 2 for i in cyc):
			#				 subs.add('cyclic')
			#				 break
			#	 if 'cyclic' not in subs:
			#		 for G in Gn:
			#			 cyc = list(nx.simple_cycles(G.to_directed()))
			#			 if any(len(i) > 2 for i in cyc):
			#				 subs.add('cyclic')
			#				 break

			return subs


	def _get_class_num(self):
		return len(set(self._targets))


	def _get_node_attr_dim(self):
		return len(self._node_attrs)


	def _get_edge_attr_dim(self):
		return len(self._edge_attrs)


	def _compute_all_degree_entropy(self, base=None):
		"""Compute the entropy of degree distribution of each graph.

		Parameters
		----------
		base : float, optional
			The logarithmic base to use. The default is ``e`` (natural logarithm).

		Returns
		-------
		degree_entropy : float
			The calculated entropy.
		"""
		from gklearn.utils.stats import entropy

		degree_entropy = []
		for g in self._graphs:
			degrees = list(dict(g.degree()).values())
			en = entropy(degrees, base=base)
			degree_entropy.append(en)
		return degree_entropy


	@property
	def graphs(self):
		return self._graphs


	@property
	def targets(self):
		return self._targets


	@property
	def node_labels(self):
		return self._node_labels


	@property
	def edge_labels(self):
		return self._edge_labels


	@property
	def node_attrs(self):
		return self._node_attrs


	@property
	def edge_attrs(self):
		return self._edge_attrs


def split_dataset_by_target(dataset):
	import warnings
	warnings.simplefilter('always', DeprecationWarning)
	warnings.warn('This function has been moved to "gklearn.dataset" module. The function "gklearn.utils.dataset.split_dataset_by_target" has not been maintained since Nov 12th, 2020 (version 0.2.1) and will be removed since version 0.4.0.', DeprecationWarning)

	from gklearn.preimage.utils import get_same_item_indices

	graphs = dataset.graphs
	targets = dataset.targets
	datasets = []
	idx_targets = get_same_item_indices(targets)
	for key, val in idx_targets.items():
		sub_graphs = [graphs[i] for i in val]
		sub_dataset = Dataset()
		sub_dataset.load_graphs(sub_graphs, [key] * len(val))
		node_labels = dataset.node_labels.copy() if dataset.node_labels is not None else None
		node_attrs = dataset.node_attrs.copy() if dataset.node_attrs is not None else None
		edge_labels = dataset.edge_labels.copy() if dataset.edge_labels is not None else None
		edge_attrs = dataset.edge_attrs.copy() if dataset.edge_attrs is not None else None
		sub_dataset.set_labels(node_labels=node_labels, node_attrs=node_attrs, edge_labels=edge_labels, edge_attrs=edge_attrs)
		datasets.append(sub_dataset)
		# @todo: clean_labels?
	return datasets


""" Utilities function to manage graph files
"""
from os.path import dirname, splitext

def loadCT(filename):
	"""load data from a Chemical Table (.ct) file.

	Notes
	------
	a typical example of data in .ct is like this:

	3 2  <- number of nodes and edges

	0.0000	0.0000	0.0000 C <- each line describes a node (x,y,z + label)

	0.0000	0.0000	0.0000 C

	0.0000	0.0000	0.0000 O

	1  3  1  1 <- each line describes an edge : to, from, bond type, bond stereo

	2  3  1  1
	  
	Check `CTFile Formats file <https://www.google.com/url?sa=t&rct=j&q=&esrc=s&source=web&cd=10&ved=2ahUKEwivhaSdjsTlAhVhx4UKHczHA8gQFjAJegQIARAC&url=https%3A%2F%2Fwww.daylight.com%2Fmeetings%2Fmug05%2FKappler%2Fctfile.pdf&usg=AOvVaw1cDNrrmMClkFPqodlF2inS>`__
	for detailed format discription.
	"""
	import networkx as nx
	from os.path import basename
	g = nx.Graph()
	with open(filename) as f:
		content = f.read().splitlines()
		g = nx.Graph(
			name = str(content[0]),
			filename = basename(filename))  # set name of the graph
		tmp = content[1].split(" ")
		if tmp[0] == '':
			nb_nodes = int(tmp[1])  # number of the nodes
			nb_edges = int(tmp[2])  # number of the edges
		else:
			nb_nodes = int(tmp[0])
			nb_edges = int(tmp[1])
			# patch for compatibility : label will be removed later
		for i in range(0, nb_nodes):
			tmp = content[i + 2].split(" ")
			tmp = [x for x in tmp if x != '']
			g.add_node(i, atom=tmp[3].strip(), 
					   label=[item.strip() for item in tmp[3:]], 
					   attributes=[item.strip() for item in tmp[0:3]])
		for i in range(0, nb_edges):
			tmp = content[i + g.number_of_nodes() + 2].split(" ")
			tmp = [x for x in tmp if x != '']
			g.add_edge(int(tmp[0]) - 1, int(tmp[1]) - 1,
					   bond_type=tmp[2].strip(),
					   label=[item.strip() for item in tmp[2:]])
	return g


def loadGXL(filename):
	from os.path import basename
	import networkx as nx
	import xml.etree.ElementTree as ET

	tree = ET.parse(filename)
	root = tree.getroot()
	index = 0
	g = nx.Graph(filename=basename(filename), name=root[0].attrib['id'])
	dic = {}  # used to retrieve incident nodes of edges
	for node in root.iter('node'):
		dic[node.attrib['id']] = index
		labels = {}
		for attr in node.iter('attr'):
			labels[attr.attrib['name']] = attr[0].text
		if 'chem' in labels:
			labels['label'] = labels['chem']
			labels['atom'] = labels['chem']
		g.add_node(index, **labels)
		index += 1

	for edge in root.iter('edge'):
		labels = {}
		for attr in edge.iter('attr'):
			labels[attr.attrib['name']] = attr[0].text
		if 'valence' in labels:
			labels['label'] = labels['valence']
			labels['bond_type'] = labels['valence']
		g.add_edge(dic[edge.attrib['from']], dic[edge.attrib['to']], **labels)
	return g


def saveGXL(graph, filename, method='default', node_labels=[], edge_labels=[], node_attrs=[], edge_attrs=[]):
	if method == 'default':
		gxl_file = open(filename, 'w')
		gxl_file.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n")
		gxl_file.write("<!DOCTYPE gxl SYSTEM \"http://www.gupro.de/GXL/gxl-1.0.dtd\">\n")
		gxl_file.write("<gxl xmlns:xlink=\"http://www.w3.org/1999/xlink\">\n")
		if 'name' in graph.graph:
			name = str(graph.graph['name'])
		else:
			name = 'dummy'
		gxl_file.write("<graph id=\"" + name + "\" edgeids=\"false\" edgemode=\"undirected\">\n")
		for v, attrs in graph.nodes(data=True):
			gxl_file.write("<node id=\"_" + str(v) + "\">")
			for l_name in node_labels:	
				gxl_file.write("<attr name=\"" + l_name + "\"><int>" + 
							   str(attrs[l_name]) + "</int></attr>")
			for a_name in node_attrs:	
				gxl_file.write("<attr name=\"" + a_name + "\"><float>" + 
							   str(attrs[a_name]) + "</float></attr>")
			gxl_file.write("</node>\n")
		for v1, v2, attrs in graph.edges(data=True):
			gxl_file.write("<edge from=\"_" + str(v1) + "\" to=\"_" + str(v2) + "\">")
			for l_name in edge_labels:	
				gxl_file.write("<attr name=\"" + l_name + "\"><int>" + 
							   str(attrs[l_name]) + "</int></attr>")
			for a_name in edge_attrs:	
				gxl_file.write("<attr name=\"" + a_name + "\"><float>" + 
							   str(attrs[a_name]) + "</float></attr>")
			gxl_file.write("</edge>\n")
		gxl_file.write("</graph>\n")
		gxl_file.write("</gxl>")
		gxl_file.close()
	elif method == 'benoit':
		import xml.etree.ElementTree as ET
		root_node = ET.Element('gxl')
		attr = dict()
		attr['id'] = str(graph.graph['name'])
		attr['edgeids'] = 'true'
		attr['edgemode'] = 'undirected'
		graph_node = ET.SubElement(root_node, 'graph', attrib=attr)
	
		for v in graph:
			current_node = ET.SubElement(graph_node, 'node', attrib={'id': str(v)})
			for attr in graph.nodes[v].keys():
				cur_attr = ET.SubElement(
					current_node, 'attr', attrib={'name': attr})
				cur_value = ET.SubElement(cur_attr,
										  graph.nodes[v][attr].__class__.__name__)
				cur_value.text = graph.nodes[v][attr]
	
		for v1 in graph:
			for v2 in graph[v1]:
				if (v1 < v2):  # Non oriented graphs
					cur_edge = ET.SubElement(
						graph_node,
						'edge',
						attrib={
							'from': str(v1),
							'to': str(v2)
						})
					for attr in graph[v1][v2].keys():
						cur_attr = ET.SubElement(
							cur_edge, 'attr', attrib={'name': attr})
						cur_value = ET.SubElement(
							cur_attr, graph[v1][v2][attr].__class__.__name__)
						cur_value.text = str(graph[v1][v2][attr])
	
		tree = ET.ElementTree(root_node)
		tree.write(filename)
	elif method == 'gedlib':
		# reference: https://github.com/dbblumenthal/gedlib/blob/master/data/generate_molecules.py#L22
#		pass
		gxl_file = open(filename, 'w')
		gxl_file.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n")
		gxl_file.write("<!DOCTYPE gxl SYSTEM \"http://www.gupro.de/GXL/gxl-1.0.dtd\">\n")
		gxl_file.write("<gxl xmlns:xlink=\"http://www.w3.org/1999/xlink\">\n")
		gxl_file.write("<graph id=\"" + str(graph.graph['name']) + "\" edgeids=\"true\" edgemode=\"undirected\">\n")
		for v, attrs in graph.nodes(data=True):
			gxl_file.write("<node id=\"_" + str(v) + "\">")
			gxl_file.write("<attr name=\"" + "chem" + "\"><int>" + str(attrs['chem']) + "</int></attr>")
			gxl_file.write("</node>\n")
		for v1, v2, attrs in graph.edges(data=True):
			gxl_file.write("<edge from=\"_" + str(v1) + "\" to=\"_" + str(v2) + "\">")
			gxl_file.write("<attr name=\"valence\"><int>" + str(attrs['valence']) + "</int></attr>")
#			gxl_file.write("<attr name=\"valence\"><int>" + "1" + "</int></attr>")
			gxl_file.write("</edge>\n")
		gxl_file.write("</graph>\n")
		gxl_file.write("</gxl>")
		gxl_file.close()
	elif method == 'gedlib-letter':
		# reference: https://github.com/dbblumenthal/gedlib/blob/master/data/generate_molecules.py#L22
		# and https://github.com/dbblumenthal/gedlib/blob/master/data/datasets/Letter/HIGH/AP1_0000.gxl
		gxl_file = open(filename, 'w')
		gxl_file.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n")
		gxl_file.write("<!DOCTYPE gxl SYSTEM \"http://www.gupro.de/GXL/gxl-1.0.dtd\">\n")
		gxl_file.write("<gxl xmlns:xlink=\"http://www.w3.org/1999/xlink\">\n")
		gxl_file.write("<graph id=\"" + str(graph.graph['name']) + "\" edgeids=\"false\" edgemode=\"undirected\">\n")
		for v, attrs in graph.nodes(data=True):
			gxl_file.write("<node id=\"_" + str(v) + "\">")
			gxl_file.write("<attr name=\"x\"><float>" + str(attrs['attributes'][0]) + "</float></attr>")
			gxl_file.write("<attr name=\"y\"><float>" + str(attrs['attributes'][1]) + "</float></attr>")
			gxl_file.write("</node>\n")
		for v1, v2, attrs in graph.edges(data=True):
			gxl_file.write("<edge from=\"_" + str(v1) + "\" to=\"_" + str(v2) + "\"/>\n")
		gxl_file.write("</graph>\n")
		gxl_file.write("</gxl>")
		gxl_file.close()


def loadSDF(filename):
	"""load data from structured data file (.sdf file).

	Notes
	------
	A SDF file contains a group of molecules, represented in the similar way as in MOL format.
	Check `here <http://www.nonlinear.com/progenesis/sdf-studio/v0.9/faq/sdf-file-format-guidance.aspx>`__ for detailed structure.
	"""
	import networkx as nx
	from os.path import basename
	from tqdm import tqdm
	import sys
	data = []
	with open(filename) as f:
		content = f.read().splitlines()
		index = 0
		pbar = tqdm(total=len(content) + 1, desc='load SDF', file=sys.stdout)
		while index < len(content):
			index_old = index

			g = nx.Graph(name=content[index].strip())  # set name of the graph

			tmp = content[index + 3]
			nb_nodes = int(tmp[:3])  # number of the nodes
			nb_edges = int(tmp[3:6])  # number of the edges

			for i in range(0, nb_nodes):
				tmp = content[i + index + 4]
				g.add_node(i, atom=tmp[31:34].strip())

			for i in range(0, nb_edges):
				tmp = content[i + index + g.number_of_nodes() + 4]
				tmp = [tmp[i:i + 3] for i in range(0, len(tmp), 3)]
				g.add_edge(
					int(tmp[0]) - 1, int(tmp[1]) - 1, bond_type=tmp[2].strip())

			data.append(g)

			index += 4 + g.number_of_nodes() + g.number_of_edges()
			while content[index].strip() != '$$$$':  # seperator
				index += 1
			index += 1

			pbar.update(index - index_old)
		pbar.update(1)
		pbar.close()

	return data


def loadMAT(filename, extra_params):
	"""Load graph data from a MATLAB (up to version 7.1) .mat file.

	Notes
	------
	A MAT file contains a struct array containing graphs, and a column vector lx containing a class label for each graph.
	Check README in `downloadable file <http://mlcb.is.tuebingen.mpg.de/Mitarbeiter/Nino/WL/>`__ for detailed structure.
	"""
	from scipy.io import loadmat
	import numpy as np
	import networkx as nx
	data = []
	content = loadmat(filename)
	order = extra_params['am_sp_al_nl_el']
	# print(content)
	# print('----')
	for key, value in content.items():
		if key[0] == 'l':  # class label
			y = np.transpose(value)[0].tolist()
			# print(y)
		elif key[0] != '_':
			# print(value[0][0][0])
			# print()
			# print(value[0][0][1])
			# print()
			# print(value[0][0][2])
			# print()
			# if len(value[0][0]) > 3:
			#	 print(value[0][0][3])
			# print('----')
			# if adjacency matrix is not compressed / edge label exists
			if order[1] == 0:
				for i, item in enumerate(value[0]):
					# print(item)
					# print('------')
					g = nx.Graph(name=i)  # set name of the graph
					nl = np.transpose(item[order[3]][0][0][0])  # node label
					# print(item[order[3]])
					# print()
					for index, label in enumerate(nl[0]):
						g.add_node(index, atom=str(label))
					el = item[order[4]][0][0][0]  # edge label
					for edge in el:
						g.add_edge(
							edge[0] - 1, edge[1] - 1, bond_type=str(edge[2]))
					data.append(g)
			else:
				from scipy.sparse import csc_matrix
				for i, item in enumerate(value[0]):
					# print(item)
					# print('------')
					g = nx.Graph(name=i)  # set name of the graph
					nl = np.transpose(item[order[3]][0][0][0])  # node label
					# print(nl)
					# print()
					for index, label in enumerate(nl[0]):
						g.add_node(index, atom=str(label))
					sam = item[order[0]]  # sparse adjacency matrix
					index_no0 = sam.nonzero()
					for col, row in zip(index_no0[0], index_no0[1]):
						# print(col)
						# print(row)
						g.add_edge(col, row)
					data.append(g)
					# print(g.edges(data=True))
	return data, y


def loadTXT(filename):
	"""Load graph data from a .txt file.

	Notes
	------
	The graph data is loaded from separate files.
	Check README in `downloadable file <http://tiny.cc/PK_MLJ_data>`__, 2018 for detailed structure.
	"""
#	import numpy as np
	import networkx as nx
	from os import listdir
	from os.path import dirname, basename
	
	
	def get_label_names(frm):
		"""Get label names from DS_label_readme.txt file.
		"""
		
		def get_names_from_line(line):
			"""Get names of labels/attributes from a line.
			"""
			str_names = line.split('[')[1].split(']')[0]
			names = str_names.split(',')
			names = [attr.strip() for attr in names]
			return names


		label_names = {'node_labels': [], 'node_attrs': [], 
					   'edge_labels': [], 'edge_attrs': []}
		content_rm = open(frm).read().splitlines()
		for line in content_rm:
			line = line.strip()
			if line.startswith('Node labels:'):
				label_names['node_labels'] = get_names_from_line(line)
			elif line.startswith('Node attributes:'):
				label_names['node_attrs'] = get_names_from_line(line)
			elif line.startswith('Edge labels:'):
				label_names['edge_labels'] = get_names_from_line(line)
			elif line.startswith('Edge attributes:'):
				label_names['edge_attrs'] = get_names_from_line(line)
		return label_names

	
	# get dataset name.
	dirname_dataset = dirname(filename)
	filename = basename(filename)
	fn_split = filename.split('_A')
	ds_name = fn_split[0].strip()
	
	# load data file names
	for name in listdir(dirname_dataset):
		if ds_name + '_A' in name:
			fam = dirname_dataset + '/' + name
		elif ds_name + '_graph_indicator' in name:
			fgi = dirname_dataset + '/' + name
		elif ds_name + '_graph_labels' in name:
			fgl = dirname_dataset + '/' + name
		elif ds_name + '_node_labels' in name:
			fnl = dirname_dataset + '/' + name
		elif ds_name + '_edge_labels' in name:
			fel = dirname_dataset + '/' + name
		elif ds_name + '_edge_attributes' in name:
			fea = dirname_dataset + '/' + name
		elif ds_name + '_node_attributes' in name:
			fna = dirname_dataset + '/' + name
		elif ds_name + '_graph_attributes' in name:
			fga = dirname_dataset + '/' + name
		elif ds_name + '_label_readme' in name:
			frm = dirname_dataset + '/' + name
		# this is supposed to be the node attrs, make sure to put this as the last 'elif'
		elif ds_name + '_attributes' in name:
			fna = dirname_dataset + '/' + name
			
	# get labels and attributes names.
	if 'frm' in locals():
		label_names = get_label_names(frm)
	else:
		label_names = {'node_labels': [], 'node_attrs': [], 
					   'edge_labels': [], 'edge_attrs': []}

	content_gi = open(fgi).read().splitlines()  # graph indicator
	content_am = open(fam).read().splitlines()  # adjacency matrix
	content_gl = open(fgl).read().splitlines()  # graph labels

	# create graphs and add nodes
	data = [nx.Graph(name=str(i), 
					 node_labels=label_names['node_labels'],
					 node_attrs=label_names['node_attrs'],
					 edge_labels=label_names['edge_labels'],
					 edge_attrs=label_names['edge_attrs']) for i in range(0, len(content_gl))]
	if 'fnl' in locals():
		content_nl = open(fnl).read().splitlines()  # node labels
		for idx, line in enumerate(content_gi):
			# transfer to int first in case of unexpected blanks
			data[int(line) - 1].add_node(idx)
			labels = [l.strip() for l in content_nl[idx].split(',')]
			data[int(line) - 1].nodes[idx]['atom'] = str(int(labels[0])) # @todo: this should be removed after.
			if data[int(line) - 1].graph['node_labels'] == []:
				for i, label in enumerate(labels):
					l_name = 'label_' + str(i)
					data[int(line) - 1].nodes[idx][l_name] = label
					data[int(line) - 1].graph['node_labels'].append(l_name)
			else:
				for i, l_name in enumerate(data[int(line) - 1].graph['node_labels']):
					data[int(line) - 1].nodes[idx][l_name] = labels[i]
	else:
		for i, line in enumerate(content_gi):
			data[int(line) - 1].add_node(i)
			
	# add edges
	for line in content_am:
		tmp = line.split(',')
		n1 = int(tmp[0]) - 1
		n2 = int(tmp[1]) - 1
		# ignore edge weight here.
		g = int(content_gi[n1]) - 1
		data[g].add_edge(n1, n2)

	# add edge labels
	if 'fel' in locals():
		content_el = open(fel).read().splitlines()
		for idx, line in enumerate(content_el):
			labels = [l.strip() for l in line.split(',')]
			n = [int(i) - 1 for i in content_am[idx].split(',')]
			g = int(content_gi[n[0]]) - 1
			data[g].edges[n[0], n[1]]['bond_type'] = labels[0] # @todo: this should be removed after.
			if data[g].graph['edge_labels'] == []:
				for i, label in enumerate(labels):
					l_name = 'label_' + str(i)
					data[g].edges[n[0], n[1]][l_name] = label
					data[g].graph['edge_labels'].append(l_name)
			else:
				for i, l_name in enumerate(data[g].graph['edge_labels']):
					data[g].edges[n[0], n[1]][l_name] = labels[i]

	# add node attributes
	if 'fna' in locals():
		content_na = open(fna).read().splitlines()
		for idx, line in enumerate(content_na):
			attrs = [a.strip() for a in line.split(',')]
			g = int(content_gi[idx]) - 1
			data[g].nodes[idx]['attributes'] = attrs # @todo: this should be removed after.
			if data[g].graph['node_attrs'] == []:
				for i, attr in enumerate(attrs):
					a_name = 'attr_' + str(i)
					data[g].nodes[idx][a_name] = attr
					data[g].graph['node_attrs'].append(a_name)
			else:
				for i, a_name in enumerate(data[g].graph['node_attrs']):
					data[g].nodes[idx][a_name] = attrs[i]

	# add edge attributes
	if 'fea' in locals():
		content_ea = open(fea).read().splitlines()
		for idx, line in enumerate(content_ea):
			attrs = [a.strip() for a in line.split(',')]
			n = [int(i) - 1 for i in content_am[idx].split(',')]
			g = int(content_gi[n[0]]) - 1
			data[g].edges[n[0], n[1]]['attributes'] = attrs # @todo: this should be removed after.
			if data[g].graph['edge_attrs'] == []:
				for i, attr in enumerate(attrs):
					a_name = 'attr_' + str(i)
					data[g].edges[n[0], n[1]][a_name] = attr
					data[g].graph['edge_attrs'].append(a_name)
			else:
				for i, a_name in enumerate(data[g].graph['edge_attrs']):
					data[g].edges[n[0], n[1]][a_name] = attrs[i]

	# load y
	y = [int(i) for i in content_gl]

	return data, y


def loadDataset(filename, filename_y=None, extra_params=None):
	"""Read graph data from filename and load them as NetworkX graphs.

	Parameters
	----------
	filename : string
		The name of the file from where the dataset is read.
	filename_y : string
		The name of file of the targets corresponding to graphs.
	extra_params : dict
		Extra parameters only designated to '.mat' format.

	Return
	------
	data : List of NetworkX graph.

	y : List

	Targets corresponding to graphs.
		
	Notes
	-----
	This function supports following graph dataset formats:

	'ds': load data from .ds file. See comments of function loadFromDS for a example.

	'cxl': load data from Graph eXchange Language file (.cxl file). See 
	`here <http://www.gupro.de/GXL/Introduction/background.html>`__ for detail.

	'sdf': load data from structured data file (.sdf file). See 
	`here <http://www.nonlinear.com/progenesis/sdf-studio/v0.9/faq/sdf-file-format-guidance.aspx>`__
	for details.

	'mat': Load graph data from a MATLAB (up to version 7.1) .mat file. See
	README in `downloadable file <http://mlcb.is.tuebingen.mpg.de/Mitarbeiter/Nino/WL/>`__
	for details.

	'txt': Load graph data from a special .txt file. See
	`here <https://ls11-www.cs.tu-dortmund.de/staff/morris/graphkerneldatasets>`__
	for details. Note here filename is the name of either .txt file in
	the dataset directory.
	"""
	extension = splitext(filename)[1][1:]
	if extension == "ds":
		data, y = loadFromDS(filename, filename_y)
	elif extension == "cxl":
		import xml.etree.ElementTree as ET
		
		dirname_dataset = dirname(filename)
		tree = ET.parse(filename)
		root = tree.getroot()
		data = []
		y = []
		for graph in root.iter('graph'):
			mol_filename = graph.attrib['file']
			mol_class = graph.attrib['class']
			data.append(loadGXL(dirname_dataset + '/' + mol_filename))
			y.append(mol_class)
	elif extension == 'xml':
		data, y = loadFromXML(filename, extra_params)
	elif extension == "sdf":
#		import numpy as np
		from tqdm import tqdm
		import sys

		data = loadSDF(filename)

		y_raw = open(filename_y).read().splitlines()
		y_raw.pop(0)
		tmp0 = []
		tmp1 = []
		for i in range(0, len(y_raw)):
			tmp = y_raw[i].split(',')
			tmp0.append(tmp[0])
			tmp1.append(tmp[1].strip())

		y = []
		for i in tqdm(range(0, len(data)), desc='ajust data', file=sys.stdout):
			try:
				y.append(tmp1[tmp0.index(data[i].name)].strip())
			except ValueError:  # if data[i].name not in tmp0
				data[i] = []
		data = list(filter(lambda a: a != [], data))
	elif extension == "mat":
		data, y = loadMAT(filename, extra_params)
	elif extension == 'txt':
		data, y = loadTXT(filename)
		# print(len(y))
		# print(y)
		# print(data[0].nodes(data=True))
		# print('----')
		# print(data[0].edges(data=True))
		# for g in data:
		#	 print(g.nodes(data=True))
		#	 print('----')
		#	 print(g.edges(data=True))

	return data, y


def loadFromXML(filename, extra_params):
	import xml.etree.ElementTree as ET
	
	if extra_params:
		dirname_dataset = extra_params
	else:
		dirname_dataset = dirname(filename)
	tree = ET.parse(filename)
	root = tree.getroot()
	data = []
	y = []
	for graph in root.iter('graph'):
		mol_filename = graph.attrib['file']
		mol_class = graph.attrib['class']
		data.append(loadGXL(dirname_dataset + '/' + mol_filename))
		y.append(mol_class)
		
	return data, y
			

def loadFromDS(filename, filename_y):
	"""Load data from .ds file.

	Possible graph formats include:

	'.ct': see function loadCT for detail.

	'.gxl': see dunction loadGXL for detail.

	Note these graph formats are checked automatically by the extensions of 
	graph files.
	"""
	dirname_dataset = dirname(filename)
	data = []
	y = []
	content = open(filename).read().splitlines()
	extension = splitext(content[0].split(' ')[0])[1][1:]
	if filename_y is None or filename_y == '':
		if extension == 'ct': 
			for i in range(0, len(content)):
				tmp = content[i].split(' ')
				# remove the '#'s in file names
				data.append(
					loadCT(dirname_dataset + '/' + tmp[0].replace('#', '', 1)))
				y.append(float(tmp[1]))
		elif extension == 'gxl':
			for i in range(0, len(content)):
				tmp = content[i].split(' ')
				# remove the '#'s in file names
				data.append(
					loadGXL(dirname_dataset + '/' + tmp[0].replace('#', '', 1)))
				y.append(float(tmp[1]))
	else:  # y in a seperate file
		if extension == 'ct':
			for i in range(0, len(content)):
				tmp = content[i]
				# remove the '#'s in file names
				data.append(
					loadCT(dirname_dataset + '/' + tmp.replace('#', '', 1)))
		elif extension == 'gxl':
			for i in range(0, len(content)):
				tmp = content[i]
				# remove the '#'s in file names
				data.append(
					loadGXL(dirname_dataset + '/' + tmp.replace('#', '', 1)))													   
															   
		content_y = open(filename_y).read().splitlines()
		# assume entries in filename and filename_y have the same order.
		for item in content_y:
			tmp = item.split(' ')
			# assume the 3rd entry in a line is y (for Alkane dataset)
			y.append(float(tmp[2]))
			
	return data, y
				

def saveDataset(Gn, y, gformat='gxl', group=None, filename='gfile', xparams=None):
	"""Save list of graphs.
	"""
	import os
	dirname_ds = os.path.dirname(filename)
	if dirname_ds != '':
		dirname_ds += '/'
		os.makedirs(dirname_ds, exist_ok=True)
				
	if xparams is not None and 'graph_dir' in xparams:
		graph_dir = xparams['graph_dir'] + '/'
		os.makedirs(graph_dir, exist_ok=True)
	else:
		graph_dir = dirname_ds 
			
	if group == 'xml' and gformat == 'gxl':
		kwargs = {'method': xparams['method']} if xparams is not None else {}
		with open(filename + '.xml', 'w') as fgroup:
			fgroup.write("<?xml version=\"1.0\"?>")
			fgroup.write("\n<!DOCTYPE GraphCollection SYSTEM \"http://www.inf.unibz.it/~blumenthal/dtd/GraphCollection.dtd\">")
			fgroup.write("\n<GraphCollection>")
			for idx, g in enumerate(Gn):
				fname_tmp = "graph" + str(idx) + ".gxl"
				saveGXL(g, graph_dir + fname_tmp, **kwargs)
				fgroup.write("\n\t<graph file=\"" + fname_tmp + "\" class=\"" + str(y[idx]) + "\"/>")
			fgroup.write("\n</GraphCollection>")
			fgroup.close()
			
			
if __name__ == '__main__':	
#	### Load dataset from .ds file.
#	# .ct files.
#	ds = {'name': 'Alkane', 'dataset': '../../datasets/Alkane/dataset.ds',
#		'dataset_y': '../../datasets/Alkane/dataset_boiling_point_names.txt'}
#	Gn, y = loadDataset(ds['dataset'], filename_y=ds['dataset_y'])
##	ds = {'name': 'Acyclic', 'dataset': '../../datasets/acyclic/dataset_bps.ds'}  # node symb
##	Gn, y = loadDataset(ds['dataset'])
##	ds = {'name': 'MAO', 'dataset': '../../datasets/MAO/dataset.ds'} # node/edge symb
##	Gn, y = loadDataset(ds['dataset'])
##	ds = {'name': 'PAH', 'dataset': '../../datasets/PAH/dataset.ds'} # unlabeled
##	Gn, y = loadDataset(ds['dataset'])
#	print(Gn[1].nodes(data=True))
#	print(Gn[1].edges(data=True))
#	print(y[1])
	
#	# .gxl file.
#	ds = {'name': 'monoterpenoides', 
#		  'dataset': '../../datasets/monoterpenoides/dataset_10+.ds'}  # node/edge symb
#	Gn, y = loadDataset(ds['dataset'])
#	print(Gn[1].nodes(data=True))
#	print(Gn[1].edges(data=True))
#	print(y[1])
	
#	### Convert graph from one format to another.
#	# .gxl file.
#	import networkx as nx
#	ds = {'name': 'monoterpenoides', 
#		  'dataset': '../../datasets/monoterpenoides/dataset_10+.ds'}  # node/edge symb
#	Gn, y = loadDataset(ds['dataset'])
#	y = [int(i) for i in y]
#	print(Gn[1].nodes(data=True))
#	print(Gn[1].edges(data=True))
#	print(y[1])
#	# Convert a graph to the proper NetworkX format that can be recognized by library gedlib.
#	Gn_new = []
#	for G in Gn:
#		G_new = nx.Graph()
#		for nd, attrs in G.nodes(data=True):
#			G_new.add_node(str(nd), chem=attrs['atom'])
#		for nd1, nd2, attrs in G.edges(data=True):
#			G_new.add_edge(str(nd1), str(nd2), valence=attrs['bond_type'])
##			G_new.add_edge(str(nd1), str(nd2))
#		Gn_new.append(G_new)
#	print(Gn_new[1].nodes(data=True))
#	print(Gn_new[1].edges(data=True))
#	print(Gn_new[1])
#	filename = '/media/ljia/DATA/research-repo/codes/others/gedlib/tests_linlin/generated_datsets/monoterpenoides/gxl/monoterpenoides'
#	xparams = {'method': 'gedlib'}
#	saveDataset(Gn, y, gformat='gxl', group='xml', filename=filename, xparams=xparams)
	
	# save dataset.
#	ds = {'name': 'MUTAG', 'dataset': '../../datasets/MUTAG/MUTAG.mat',
#		  'extra_params': {'am_sp_al_nl_el': [0, 0, 3, 1, 2]}}  # node/edge symb
#	Gn, y = loadDataset(ds['dataset'], extra_params=ds['extra_params'])
#	saveDataset(Gn, y, group='xml', filename='temp/temp')
	
	# test - new way to add labels and attributes.
#	dataset = '../../datasets/SYNTHETICnew/SYNTHETICnew_A.txt'
#	dataset = '../../datasets/Fingerprint/Fingerprint_A.txt'
#	dataset = '../../datasets/Letter-med/Letter-med_A.txt'
#	dataset = '../../datasets/AIDS/AIDS_A.txt'
#	dataset = '../../datasets/ENZYMES_txt/ENZYMES_A_sparse.txt'
#	Gn, y_all = loadDataset(dataset)
	pass


"""
median

Methods to compute the median of a set of graphs.

@Author: linlin
@Date: 03.05.23
"""
import networkx as nx


def set_median_graph(graphs, dist_mat=None):
	"""Compute the set median of a set of graphs.

	Parameters
	----------
	graphs : list of networkx.Graph
		A set of graphs.
	dist_mat : numpy.matrix, optional
		The distance matrix of the graphs. If None, it will be computed using
		networkx.graph_edit_distance.

	Returns
	-------
	networkx.Graph
		The set median of the graphs.
	int
		The index of the set median in the input list.

	Authors
	-------
	Linlin Jia, Github Copilot (2023.05.03)
	"""
	# Compute the median of a set of graphs using Graph Edit Distance.
	if len(graphs) == 1:
		return graphs[0], 0

	if dist_mat is None:
		dist_mat = [[0 for _ in range(len(graphs))] for _ in range(len(graphs))]
		for i in range(len(graphs)):
			for j in range(i + 1, len(graphs)):
				dist_mat[i][j] = nx.graph_edit_distance(graphs[i], graphs[j])
				dist_mat[j][i] = dist_mat[i][j]
	else:
		if len(graphs) != len(dist_mat):
			raise ValueError(
				'The number of graphs and the distance matrix do not match.'
			)
	# print(dist_mat) # @TODO: remove
	dist_vec = [sum(v) for v in dist_mat]

	idx = dist_vec.index(min(dist_vec))
	set_median = graphs[idx]

	return set_median, idx



from ctypes import *
import os
lib1 = cdll.LoadLibrary(os.path.dirname(os.path.realpath(__file__)) + '/lib/fann.2.2.0/libdoublefann.so')
lib2 = cdll.LoadLibrary(os.path.dirname(os.path.realpath(__file__)) + '/lib/libsvm.3.22/libsvm.so')
lib3 = cdll.LoadLibrary(os.path.dirname(os.path.realpath(__file__)) + '/lib/nomad.3.8.1/libnomad.so')
lib4 = cdll.LoadLibrary(os.path.dirname(os.path.realpath(__file__)) + '/lib/nomad.3.8.1/libsgtelib.so')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 16 16:04:46 2020

@author: ljia
"""
import numpy as np
import time
import random
import multiprocessing
import networkx as nx
from gklearn.preimage import PreimageGenerator
from gklearn.preimage.utils import compute_k_dis
from gklearn.ged.env import GEDEnv
from gklearn.ged.learning import CostMatricesLearner
from gklearn.ged.median import MedianGraphEstimatorCML
from gklearn.ged.median import constant_node_costs, mge_options_to_string
from gklearn.utils.utils import get_graph_kernel_by_name
from gklearn.ged.util import label_costs_to_matrix


class MedianPreimageGeneratorCML(PreimageGenerator):
	"""Generator median preimages by cost matrices learning using the pure Python version of GEDEnv. Works only for symbolic labeled graphs.
	"""
	
	def __init__(self, dataset=None):
		PreimageGenerator.__init__(self, dataset=dataset)
		### arguments to set.
		self._mge = None
		self._ged_options = {}
		self._mge_options = {}
# 		self._fit_method = 'k-graphs'
		self._init_method = 'random'
		self._init_ecc = None
		self._parallel = True
		self._n_jobs = multiprocessing.cpu_count()
		self._ds_name = None
		# for cml.
		self._time_limit_in_sec = 0
		self._max_itrs = 100
		self._max_itrs_without_update = 3
		self._epsilon_residual = 0.01
		self._epsilon_ec = 0.1
		self._allow_zeros = True
# 		self._triangle_rule = True
		### values to compute.
		self._runtime_optimize_ec = None
		self._runtime_generate_preimage = None
		self._runtime_total = None
		self._set_median = None
		self._gen_median = None
		self._best_from_dataset = None
		self._sod_set_median = None
		self._sod_gen_median = None
		self._k_dis_set_median = None
		self._k_dis_gen_median = None
		self._k_dis_dataset = None
		self._node_label_costs = None
		self._edge_label_costs = None
		# for cml.
		self._itrs = 0
		self._converged = False
		self._num_updates_ecs = 0
		### values that can be set or to be computed.
		self._edit_cost_constants = []
		self._gram_matrix_unnorm = None
		self._runtime_precompute_gm = None

		
	def set_options(self, **kwargs):
		self._kernel_options = kwargs.get('kernel_options', {})
		self._graph_kernel = kwargs.get('graph_kernel', None)
		self._verbose = kwargs.get('verbose', 2)
		self._ged_options = kwargs.get('ged_options', {})
		self._mge_options = kwargs.get('mge_options', {})
# 		self._fit_method = kwargs.get('fit_method', 'k-graphs')
		self._init_method = kwargs.get('init_method', 'random')
		self._init_ecc = kwargs.get('init_ecc', None)
		self._edit_cost_constants = kwargs.get('edit_cost_constants', [])
		self._parallel = kwargs.get('parallel', True)
		self._n_jobs = kwargs.get('n_jobs', multiprocessing.cpu_count())
		self._ds_name = kwargs.get('ds_name', None)
		self._time_limit_in_sec = kwargs.get('time_limit_in_sec', 0)
		self._max_itrs = kwargs.get('max_itrs', 100)
		self._max_itrs_without_update = kwargs.get('max_itrs_without_update', 3)
		self._epsilon_residual = kwargs.get('epsilon_residual', 0.01)
		self._epsilon_ec = kwargs.get('epsilon_ec', 0.1)
		self._gram_matrix_unnorm = kwargs.get('gram_matrix_unnorm', None)
		self._runtime_precompute_gm = kwargs.get('runtime_precompute_gm', None)
		self._allow_zeros = kwargs.get('allow_zeros', True)
# 		self._triangle_rule = kwargs.get('triangle_rule', True)
		
		
	def run(self):
		self._graph_kernel = get_graph_kernel_by_name(self._kernel_options['name'], 
						  node_labels=self._dataset.node_labels,
						  edge_labels=self._dataset.edge_labels, 
						  node_attrs=self._dataset.node_attrs,
						  edge_attrs=self._dataset.edge_attrs,
						  ds_infos=self._dataset.get_dataset_infos(keys=['directed']),
						  kernel_options=self._kernel_options)
		
		# record start time.
		start = time.time()
		
		# 1. precompute gram matrix.
		if self._gram_matrix_unnorm is None:
			gram_matrix, run_time = self._graph_kernel.compute(self._dataset.graphs, **self._kernel_options)
			self._gram_matrix_unnorm = self._graph_kernel.gram_matrix_unnorm
			end_precompute_gm = time.time()
			self._runtime_precompute_gm = end_precompute_gm - start
		else:
			if self._runtime_precompute_gm is None:
				raise Exception('Parameter "runtime_precompute_gm" must be given when using pre-computed Gram matrix.')
			self._graph_kernel.gram_matrix_unnorm = self._gram_matrix_unnorm
			if self._kernel_options['normalize']:
				self._graph_kernel.gram_matrix = self._graph_kernel.normalize_gm(np.copy(self._gram_matrix_unnorm))
			else:
				self._graph_kernel.gram_matrix = np.copy(self._gram_matrix_unnorm)
			end_precompute_gm = time.time()
			start -= self._runtime_precompute_gm
			
# 		if self._fit_method != 'k-graphs' and self._fit_method != 'whole-dataset':
# 			start = time.time()
# 			self._runtime_precompute_gm = 0
# 			end_precompute_gm = start
		
		# 2. optimize edit cost constants. 
		self._optimize_edit_cost_vector()
		end_optimize_ec = time.time()
		self._runtime_optimize_ec = end_optimize_ec - end_precompute_gm
		
		# 3. compute set median and gen median using optimized edit costs.
		if self._verbose >= 2:
			print('\nstart computing set median and gen median using optimized edit costs...\n')
		self._gmg_bcu()
		end_generate_preimage = time.time()
		self._runtime_generate_preimage = end_generate_preimage - end_optimize_ec
		self._runtime_total = end_generate_preimage - start
		if self._verbose >= 2:
			print('medians computed.')
			print('SOD of the set median: ', self._sod_set_median)
			print('SOD of the generalized median: ', self._sod_gen_median)
			
		# 4. compute kernel distances to the true median.
		if self._verbose >= 2:
			print('\nstart computing distances to true median....\n')
		self._compute_distances_to_true_median()

		# 5. print out results.
		if self._verbose:
			print()
			print('================================================================================')
			print('Finished generation of preimages.')
			print('--------------------------------------------------------------------------------')
			print('The optimized edit costs:', self._edit_cost_constants)
			print('SOD of the set median:', self._sod_set_median)
			print('SOD of the generalized median:', self._sod_gen_median)
			print('Distance in kernel space for set median:', self._k_dis_set_median)
			print('Distance in kernel space for generalized median:', self._k_dis_gen_median)
			print('Minimum distance in kernel space for each graph in median set:', self._k_dis_dataset)
			print('Time to pre-compute Gram matrix:', self._runtime_precompute_gm)
			print('Time to optimize edit costs:', self._runtime_optimize_ec)
			print('Time to generate pre-images:', self._runtime_generate_preimage)
			print('Total time:', self._runtime_total)
			print('Total number of iterations for optimizing:', self._itrs)
			print('Total number of updating edit costs:', self._num_updates_ecs)
			print('Is optimization of edit costs converged:', self._converged)
			print('================================================================================')
			print()


	def get_results(self):
		results = {}
		results['edit_cost_constants'] = self._edit_cost_constants
		results['runtime_precompute_gm'] = self._runtime_precompute_gm
		results['runtime_optimize_ec'] = self._runtime_optimize_ec
		results['runtime_generate_preimage'] = self._runtime_generate_preimage
		results['runtime_total'] = self._runtime_total
		results['sod_set_median'] = self._sod_set_median
		results['sod_gen_median'] = self._sod_gen_median
		results['k_dis_set_median'] = self._k_dis_set_median
		results['k_dis_gen_median'] = self._k_dis_gen_median
		results['k_dis_dataset'] = self._k_dis_dataset
		results['itrs'] = self._itrs
		results['converged'] = self._converged
		results['num_updates_ecc'] = self._num_updates_ecs
		results['mge'] = {}
		results['mge']['num_decrease_order'] = self._mge.get_num_times_order_decreased()
		results['mge']['num_increase_order'] = self._mge.get_num_times_order_increased()
		results['mge']['num_converged_descents'] = self._mge.get_num_converged_descents()
		return results

		
	def _optimize_edit_cost_vector(self):
		"""Learn edit cost vector.	
		"""
		 # Initialize label costs randomly.
		if self._init_method == 'random':
			# Initialize label costs.
			self._initialize_label_costs()
				
			# Optimize edit cost matrices.
			self._optimize_ecm_by_kernel_distances()
		# Initialize all label costs with the same value.
		elif self._init_method == 'uniform': # random
			pass
	
		elif self._fit_method == 'random': # random
			if self._ged_options['edit_cost'] == 'LETTER':
				self._edit_cost_constants = random.sample(range(1, 1000), 3)
				self._edit_cost_constants = [item * 0.001 for item in self._edit_cost_constants]
			elif self._ged_options['edit_cost'] == 'LETTER2':
				random.seed(time.time())
				self._edit_cost_constants = random.sample(range(1, 1000), 5)
				self._edit_cost_constants = [item * 0.01 for item in self._edit_cost_constants]
			elif self._ged_options['edit_cost'] == 'NON_SYMBOLIC':
				self._edit_cost_constants = random.sample(range(1, 1000), 6)
				self._edit_cost_constants = [item * 0.01 for item in self._edit_cost_constants]
				if self._dataset.node_attrs == []:
					self._edit_cost_constants[2] = 0
				if self._dataset.edge_attrs == []:
					self._edit_cost_constants[5] = 0
			else:
				self._edit_cost_constants = random.sample(range(1, 1000), 6)
				self._edit_cost_constants = [item * 0.01 for item in self._edit_cost_constants]
			if self._verbose >= 2:
				print('edit cost constants used:', self._edit_cost_constants)
		elif self._fit_method == 'expert': # expert
			if self._init_ecc is None:
				if self._ged_options['edit_cost'] == 'LETTER':
					self._edit_cost_constants = [0.9, 1.7, 0.75] 
				elif self._ged_options['edit_cost'] == 'LETTER2':
					self._edit_cost_constants = [0.675, 0.675, 0.75, 0.425, 0.425]
				else:
					self._edit_cost_constants = [3, 3, 1, 3, 3, 1] 
			else:
				self._edit_cost_constants = self._init_ecc
		elif self._fit_method == 'k-graphs':
			if self._init_ecc is None:
				if self._ged_options['edit_cost'] == 'LETTER':
					self._init_ecc = [0.9, 1.7, 0.75] 
				elif self._ged_options['edit_cost'] == 'LETTER2':
					self._init_ecc = [0.675, 0.675, 0.75, 0.425, 0.425]
				elif self._ged_options['edit_cost'] == 'NON_SYMBOLIC':
					self._init_ecc = [0, 0, 1, 1, 1, 0]
					if self._dataset.node_attrs == []:
						self._init_ecc[2] = 0
					if self._dataset.edge_attrs == []:
						self._init_ecc[5] = 0
				else:
					self._init_ecc = [3, 3, 1, 3, 3, 1] 
			# optimize on the k-graph subset.
			self._optimize_ecm_by_kernel_distances()
		elif self._fit_method == 'whole-dataset':
			if self._init_ecc is None:
				if self._ged_options['edit_cost'] == 'LETTER':
					self._init_ecc = [0.9, 1.7, 0.75] 
				elif self._ged_options['edit_cost'] == 'LETTER2':
					self._init_ecc = [0.675, 0.675, 0.75, 0.425, 0.425]
				else:
					self._init_ecc = [3, 3, 1, 3, 3, 1] 
			# optimizeon the whole set.
			self._optimize_ecc_by_kernel_distances()
		elif self._fit_method == 'precomputed':
			pass
		
		
	def _initialize_label_costs(self):
		self._initialize_node_label_costs()
		self._initialize_edge_label_costs()
				
				
	def _initialize_node_label_costs(self):
		# Get list of node labels.
		nls = self._dataset.get_all_node_labels()
		# Generate random costs.
		nb_nl = int((len(nls) * (len(nls) - 1)) / 2 + 2 * len(nls))
		rand_costs = random.sample(range(1, 10 * nb_nl + 1), nb_nl)
		rand_costs /= np.max(rand_costs) # @todo: maybe not needed.
		self._node_label_costs = rand_costs


	def _initialize_edge_label_costs(self):
		# Get list of edge labels.
		els = self._dataset.get_all_edge_labels()
		# Generate random costs.
		nb_el = int((len(els) * (len(els) - 1)) / 2 + 2 * len(els))
		rand_costs = random.sample(range(1, 10 * nb_el + 1), nb_el)
		rand_costs /= np.max(rand_costs) # @todo: maybe not needed.
		self._edge_label_costs = rand_costs
		
		
	def _optimize_ecm_by_kernel_distances(self):		
		# compute distances in feature space.
		dis_k_mat, _, _, _ = self._graph_kernel.compute_distance_matrix()
		dis_k_vec = []
		for i in range(len(dis_k_mat)):
	#		for j in range(i, len(dis_k_mat)):
			for j in range(i + 1, len(dis_k_mat)):
				dis_k_vec.append(dis_k_mat[i, j])
		dis_k_vec = np.array(dis_k_vec)
		
		# Set GEDEnv options.
# 		graphs = [self._clean_graph(g) for g in self._dataset.graphs]
# 		self._edit_cost_constants = self._init_ecc
		options = self._ged_options.copy()
		options['edit_cost_constants'] = self._edit_cost_constants # @todo: not needed.
		options['node_labels'] = self._dataset.node_labels
		options['edge_labels'] = self._dataset.edge_labels
# 		options['node_attrs'] = self._dataset.node_attrs
# 		options['edge_attrs'] = self._dataset.edge_attrs
		options['node_label_costs'] = self._node_label_costs
		options['edge_label_costs'] = self._edge_label_costs
		
		# Learner cost matrices.
		# Initialize cost learner.
		cml = CostMatricesLearner(edit_cost='CONSTANT', triangle_rule=False, allow_zeros=True, parallel=self._parallel, verbose=self._verbose) # @todo
		cml.set_update_params(time_limit_in_sec=self._time_limit_in_sec, max_itrs=self._max_itrs, max_itrs_without_update=self._max_itrs_without_update, epsilon_residual=self._epsilon_residual, epsilon_ec=self._epsilon_ec)
		# Run cost learner.
		cml.update(dis_k_vec, self._dataset.graphs, options)
		
		# Get results.
		results = cml.get_results()
		self._converged = results['converged']
		self._itrs = results['itrs']
		self._num_updates_ecs = results['num_updates_ecs']
		cost_list = results['cost_list']
		self._node_label_costs = cost_list[-1][0:len(self._node_label_costs)]
		self._edge_label_costs = cost_list[-1][len(self._node_label_costs):]

	
	def _gmg_bcu(self):
		"""
		The local search algorithm based on block coordinate update (BCU) for estimating a generalized median graph (GMG).

		Returns
		-------
		None.

		"""
		# Set up the ged environment.
		ged_env = GEDEnv() # @todo: maybe create a ged_env as a private varible.
		# gedlibpy.restart_env()
		ged_env.set_edit_cost(self._ged_options['edit_cost'], edit_cost_constants=self._edit_cost_constants)
		graphs = [self._clean_graph(g) for g in self._dataset.graphs]
		for g in graphs:
			ged_env.add_nx_graph(g, '')
		graph_ids = ged_env.get_all_graph_ids()
	
		node_labels = ged_env.get_all_node_labels()
		edge_labels =  ged_env.get_all_edge_labels()
		node_label_costs = label_costs_to_matrix(self._node_label_costs, len(node_labels))
		edge_label_costs = label_costs_to_matrix(self._edge_label_costs, len(edge_labels))
		ged_env.set_label_costs(node_label_costs, edge_label_costs)
	
		set_median_id = ged_env.add_graph('set_median')
		gen_median_id = ged_env.add_graph('gen_median')
		ged_env.init(init_type=self._ged_options['init_option'])
		
		# Set up the madian graph estimator.
		self._mge = MedianGraphEstimatorCML(ged_env, constant_node_costs(self._ged_options['edit_cost']))
		self._mge.set_refine_method(self._ged_options['method'], self._ged_options)
		options = self._mge_options.copy()
		if not 'seed' in options:
			options['seed'] = int(round(time.time() * 1000)) # @todo: may not work correctly for possible parallel usage.
		options['parallel'] = self._parallel
		
		# Select the GED algorithm.
		self._mge.set_options(mge_options_to_string(options))
		self._mge.set_label_names(node_labels=self._dataset.node_labels, 
					  edge_labels=self._dataset.edge_labels, 
					  node_attrs=self._dataset.node_attrs, 
					  edge_attrs=self._dataset.edge_attrs)
		ged_options = self._ged_options.copy()
		if self._parallel:
			ged_options['threads'] = 1
		self._mge.set_init_method(ged_options['method'], ged_options)
		self._mge.set_descent_method(ged_options['method'], ged_options)
		
		# Run the estimator.
		self._mge.run(graph_ids, set_median_id, gen_median_id)
		
		# Get SODs.
		self._sod_set_median = self._mge.get_sum_of_distances('initialized')
		self._sod_gen_median = self._mge.get_sum_of_distances('converged')
		
		# Get median graphs.
		self._set_median = ged_env.get_nx_graph(set_median_id)
		self._gen_median = ged_env.get_nx_graph(gen_median_id)
		
		
	def _compute_distances_to_true_median(self):		
		# compute distance in kernel space for set median.
		kernels_to_sm, _ = self._graph_kernel.compute(self._set_median, self._dataset.graphs, **self._kernel_options)
		kernel_sm, _ = self._graph_kernel.compute(self._set_median, self._set_median, **self._kernel_options)
		if self._kernel_options['normalize']:
			kernels_to_sm = [kernels_to_sm[i] / np.sqrt(self._gram_matrix_unnorm[i, i] * kernel_sm) for i in range(len(kernels_to_sm))] # normalize 
			kernel_sm = 1
		# @todo: not correct kernel value
		gram_with_sm = np.concatenate((np.array([kernels_to_sm]), np.copy(self._graph_kernel.gram_matrix)), axis=0)
		gram_with_sm = np.concatenate((np.array([[kernel_sm] + kernels_to_sm]).T, gram_with_sm), axis=1)
		self._k_dis_set_median = compute_k_dis(0, range(1, 1+len(self._dataset.graphs)), 
										  [1 / len(self._dataset.graphs)] * len(self._dataset.graphs),
										  gram_with_sm, withterm3=False)
		
		# compute distance in kernel space for generalized median.
		kernels_to_gm, _ = self._graph_kernel.compute(self._gen_median, self._dataset.graphs, **self._kernel_options)
		kernel_gm, _ = self._graph_kernel.compute(self._gen_median, self._gen_median, **self._kernel_options)
		if self._kernel_options['normalize']:
			kernels_to_gm = [kernels_to_gm[i] / np.sqrt(self._gram_matrix_unnorm[i, i] * kernel_gm) for i in range(len(kernels_to_gm))] # normalize
			kernel_gm = 1
		gram_with_gm = np.concatenate((np.array([kernels_to_gm]), np.copy(self._graph_kernel.gram_matrix)), axis=0)
		gram_with_gm = np.concatenate((np.array([[kernel_gm] + kernels_to_gm]).T, gram_with_gm), axis=1)
		self._k_dis_gen_median = compute_k_dis(0, range(1, 1+len(self._dataset.graphs)), 
										  [1 / len(self._dataset.graphs)] * len(self._dataset.graphs),
										  gram_with_gm, withterm3=False)
				
		# compute distance in kernel space for each graph in median set.
		k_dis_median_set = []
		for idx in range(len(self._dataset.graphs)):
			k_dis_median_set.append(compute_k_dis(idx+1, range(1, 1+len(self._dataset.graphs)), 
								 [1 / len(self._dataset.graphs)] * len(self._dataset.graphs), 
								 gram_with_gm, withterm3=False))
		idx_k_dis_median_set_min = np.argmin(k_dis_median_set)
		self._k_dis_dataset = k_dis_median_set[idx_k_dis_median_set_min]
		self._best_from_dataset = self._dataset.graphs[idx_k_dis_median_set_min].copy()
			
		if self._verbose >= 2:
			print()
			print('distance in kernel space for set median:', self._k_dis_set_median)
			print('distance in kernel space for generalized median:', self._k_dis_gen_median)
			print('minimum distance in kernel space for each graph in median set:', self._k_dis_dataset)
			print('distance in kernel space for each graph in median set:', k_dis_median_set)	
			
			
# 	def _clean_graph(self, G, node_labels=[], edge_labels=[], node_attrs=[], edge_attrs=[]):
	def _clean_graph(self, G): # @todo: this may not be needed when datafile is updated.
		"""
		Cleans node and edge labels and attributes of the given graph.
		"""
		G_new = nx.Graph(**G.graph)
		for nd, attrs in G.nodes(data=True):
			G_new.add_node(str(nd)) # @todo: should we keep this as str()?
			for l_name in self._dataset.node_labels:
				G_new.nodes[str(nd)][l_name] = str(attrs[l_name])
			for a_name in self._dataset.node_attrs:
				G_new.nodes[str(nd)][a_name] = str(attrs[a_name])
		for nd1, nd2, attrs in G.edges(data=True):
			G_new.add_edge(str(nd1), str(nd2))
			for l_name in self._dataset.edge_labels:
				G_new.edges[str(nd1), str(nd2)][l_name] = str(attrs[l_name])		
			for a_name in self._dataset.edge_attrs:
				G_new.edges[str(nd1), str(nd2)][a_name] = str(attrs[a_name])		
		return G_new
			
			
	@property
	def mge(self):
		return self._mge
	
	@property
	def ged_options(self):
		return self._ged_options

	@ged_options.setter
	def ged_options(self, value):
		self._ged_options = value		

	
	@property
	def mge_options(self):
		return self._mge_options

	@mge_options.setter
	def mge_options(self, value):
		self._mge_options = value		


	@property
	def fit_method(self):
		return self._fit_method

	@fit_method.setter
	def fit_method(self, value):
		self._fit_method = value
		
		
	@property
	def init_ecc(self):
		return self._init_ecc

	@init_ecc.setter
	def init_ecc(self, value):
		self._init_ecc = value
		
	
	@property
	def set_median(self):
		return self._set_median


	@property
	def gen_median(self):
		return self._gen_median
	
	
	@property
	def best_from_dataset(self):
		return self._best_from_dataset
	
	
	@property
	def gram_matrix_unnorm(self):
		return self._gram_matrix_unnorm
	
	@gram_matrix_unnorm.setter
	def gram_matrix_unnorm(self, value):
		self._gram_matrix_unnorm = value


""" Functions for python system.
"""

def isNotebook():
    """check if code is executed in the IPython notebook.
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 24 11:13:26 2022

@author: ljia

Reference: scikit-learn.
"""
from abc import abstractmethod
import numbers
import warnings
import numpy as np
from sklearn.utils import check_random_state, check_array, column_or_1d, indexable
from sklearn.utils.validation import _num_samples
from sklearn.utils.multiclass import type_of_target


class BaseCrossValidatorWithValid(object):
	"""Base class for all cross-validators.
	Implementations must define `_iter_valid_test_masks` or `_iter_valid_stest_indices`.
	"""

	def split(self, X, y=None, groups=None):
		"""Generate indices to split data into training, valid, and test set.

		Parameters
		----------

		X : array-like of shape (n_samples, n_features)
			Training data, where `n_samples` is the number of samples
			and `n_features` is the number of features.

		y : array-like of shape (n_samples,)
			The target variable for supervised learning problems.

		groups : array-like of shape (n_samples,), default=None
			Group labels for the samples used while splitting the dataset into
			train/test set.

		Yields
		------
		train : ndarray
			The training set indices for that split.

		valid : ndarray
			The valid set indices for that split.

		test : ndarray
			The testing set indices for that split.
		"""
		X, y, groups = indexable(X, y, groups)
		indices = np.arange(_num_samples(X))
		for valid_index, test_index in self._iter_valid_test_masks(X, y, groups):
			train_index = indices[np.logical_not(np.logical_or(valid_index, test_index))]
			valid_index = indices[valid_index]
			test_index = indices[test_index]
			yield train_index, valid_index, test_index


	# Since subclasses must implement either _iter_valid_test_masks or
	# _iter_valid_test_indices, neither can be abstract.
	def _iter_valid_test_masks(self, X=None, y=None, groups=None):
		"""Generates boolean masks corresponding to valid and test sets.
		By default, delegates to _iter_valid_test_indices(X, y, groups)
		"""
		for valid_index, test_index in self._iter_valid_test_indices(X, y, groups):
			valid_mask = np.zeros(_num_samples(X), dtype=bool)
			test_mask = np.zeros(_num_samples(X), dtype=bool)
			valid_mask[valid_index] = True
			test_mask[test_index] = True
			yield valid_mask, test_mask


	def _iter_valid_test_indices(self, X=None, y=None, groups=None):
		"""Generates integer indices corresponding to valid and test sets."""
		raise NotImplementedError


	@abstractmethod
	def get_n_splits(self, X=None, y=None, groups=None):
		"""Returns the number of splitting iterations in the cross-validator"""


	def __repr__(self):
		return _build_repr(self)


class _BaseKFoldWithValid(BaseCrossValidatorWithValid):
	"""Base class for KFoldWithValid, GroupKFoldWithValid, and StratifiedKFoldWithValid"""

	@abstractmethod
	def __init__(self, n_splits, *, stratify, shuffle, random_state):
		if not isinstance(n_splits, numbers.Integral):
			raise ValueError(
				'The number of folds must be of Integral type. '
				'%s of type %s was passed.' % (n_splits, type(n_splits))
			)
		n_splits = int(n_splits)

		if n_splits <= 2:
			raise ValueError(
				'k-fold cross-validation requires at least one'
				' train/valid/test split by setting n_splits=3 or more,'
				' got n_splits={0}.'.format(n_splits)
			)

		if not isinstance(shuffle, bool):
			raise TypeError('shuffle must be True or False; got {0}'.format(shuffle))

		if not shuffle and random_state is not None:  # None is the default
			raise ValueError(
				'Setting a random_state has no effect since shuffle is '
				'False. You should leave '
				'random_state to its default (None), or set shuffle=True.',
			)

		self.n_splits = n_splits
		self.stratify = stratify
		self.shuffle = shuffle
		self.random_state = random_state


	def split(self, X, y=None, groups=None):
		"""Generate indices to split data into training, valid and test set."""
		X, y, groups = indexable(X, y, groups)
		n_samples = _num_samples(X)
		if self.n_splits > n_samples:
			raise ValueError(
				(
				 'Cannot have number of splits n_splits={0} greater'
				 ' than the number of samples: n_samples={1}.'
				 ).format(self.n_splits, n_samples)
			)

		for train, valid, test in super().split(X, y, groups):
			yield train, valid, test


class KFoldWithValid(_BaseKFoldWithValid):


	def __init__(
			self,
			n_splits=5,
			*,
			stratify=False,
			shuffle=False,
			random_state=None
			):
		super().__init__(
			n_splits=n_splits,
			stratify=stratify,
			shuffle=shuffle,
			random_state=random_state
			)


	def _make_valid_test_folds(self, X, y=None):
		rng = check_random_state(self.random_state)
		y = np.asarray(y)
		type_of_target_y = type_of_target(y)
		allowed_target_types = ('binary', 'multiclass')
		if type_of_target_y not in allowed_target_types:
			raise ValueError(
				'Supported target types are: {}. Got {!r} instead.'.format(
					allowed_target_types, type_of_target_y
				)
			)

		y = column_or_1d(y)

		_, y_idx, y_inv = np.unique(y, return_index=True, return_inverse=True)
		# y_inv encodes y according to lexicographic order. We invert y_idx to
		# map the classes so that they are encoded by order of appearance:
		# 0 represents the first label appearing in y, 1 the second, etc.
		_, class_perm = np.unique(y_idx, return_inverse=True)
		y_encoded = class_perm[y_inv]

		n_classes = len(y_idx)
		y_counts = np.bincount(y_encoded)
		min_groups = np.min(y_counts)
		if np.all(self.n_splits > y_counts):
			raise ValueError(
				"n_splits=%d cannot be greater than the"
				" number of members in each class." % (self.n_splits)
			)
		if self.n_splits > min_groups:
			warnings.warn(
				"The least populated class in y has only %d"
				" members, which is less than n_splits=%d."
				% (min_groups, self.n_splits),
				UserWarning,
			)

		# Determine the optimal number of samples from each class in each fold,
		# using round robin over the sorted y. (This can be done direct from
		# counts, but that code is unreadable.)
		y_order = np.sort(y_encoded)
		allocation = np.asarray(
			[
				np.bincount(y_order[i :: self.n_splits], minlength=n_classes)
				for i in range(self.n_splits)
			]
		)

		# To maintain the data order dependencies as best as possible within
		# the stratification constraint, we assign samples from each class in
		# blocks (and then mess that up when shuffle=True).
		test_folds = np.empty(len(y), dtype='i')
		for k in range(n_classes):
			# since the kth column of allocation stores the number of samples
			# of class k in each test set, this generates blocks of fold
			# indices corresponding to the allocation for class k.
			folds_for_class = np.arange(self.n_splits).repeat(allocation[:, k])
			if self.shuffle:
				rng.shuffle(folds_for_class)
			test_folds[y_encoded == k] = folds_for_class
		return test_folds


	def _iter_valid_test_masks(self, X, y=None, groups=None):
		test_folds = self._make_valid_test_folds(X, y)
		for i in range(self.n_splits):
			if i + 1 < self.n_splits:
				j = i + 1
			else:
				j = 0
			yield test_folds == i, test_folds == j


	def split(self, X, y, groups=None):
		y = check_array(y, input_name='y', ensure_2d=False, dtype=None)
		return super().split(X, y, groups)


class _RepeatedSplitsWithValid(object):


	def __init__(
			self,
			cv,
			*,
			n_repeats=10,
			random_state=None,
			**cvargs
			):
		if not isinstance(n_repeats, int):
			raise ValueError('Number of repetitions must be of integer type.')

		if n_repeats <= 0:
			raise ValueError('Number of repetitions must be greater than 0.')

		self.cv = cv
		self.n_repeats = n_repeats
		self.random_state = random_state
		self.cvargs = cvargs


	def split(self, X, y=None, groups=None):
		n_repeats = self.n_repeats
		rng = check_random_state(self.random_state)

		for idx in range(n_repeats):
			cv = self.cv(random_state=rng, shuffle=True, **self.cvargs)
			for train_index, valid_index, test_index in cv.split(X, y, groups):
				yield train_index, valid_index, test_index


class RepeatedKFoldWithValid(_RepeatedSplitsWithValid):


	def __init__(
			self,
			*,
			n_splits=5,
			n_repeats=10,
			stratify=False,
			random_state=None
			):
		super().__init__(
			KFoldWithValid,
			n_repeats=n_repeats,
			stratify=stratify,
			random_state=random_state,
			n_splits=n_splits,
			)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct  5 15:12:41 2020

@author: ljia
"""
from collections import Counter
from scipy import stats


def entropy(labels, base=None):
	"""Calculate the entropy of a distribution for given list of labels.

	Parameters
	----------
	labels : list
		Given list of labels.
	base : float, optional
		The logarithmic base to use. The default is ``e`` (natural logarithm).

	Returns
	-------
	float
		The calculated entropy.
	"""
	return stats.entropy(list(Counter(labels).values()), base=base)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug 19 16:55:17 2020

@author: ljia

@references: 

	[1] S Vichy N Vishwanathan, Nicol N Schraudolph, Risi Kondor, and Karsten M Borgwardt. Graph kernels. Journal of Machine Learning Research, 11(Apr):1201–1242, 2010.
"""

import networkx as nx
from gklearn.utils import SpecialLabel
from gklearn.kernels import GraphKernel


class RandomWalkMeta(GraphKernel):
	
	
	def __init__(self, **kwargs):
		GraphKernel.__init__(self)
		self._weight = kwargs.get('weight', 1)
		self._p = kwargs.get('p', None)
		self._q = kwargs.get('q', None)
		self._edge_weight = kwargs.get('edge_weight', None)
		self._ds_infos = kwargs.get('ds_infos', {})
		
		
	def _compute_gm_series(self):
		pass


	def _compute_gm_imap_unordered(self):
		pass
	
		
	def _compute_kernel_list_series(self, g1, g_list):
		pass

	
	def _compute_kernel_list_imap_unordered(self, g1, g_list):
		pass
	
	
	def _compute_single_kernel_series(self, g1, g2):
		pass
	
	
	def _check_graphs(self, Gn):
		# remove graphs with no edges, as no walk can be found in their structures, 
		# so the weight matrix between such a graph and itself might be zero.
		for g in Gn:
			if nx.number_of_edges(g) == 0:
				raise Exception('Graphs must contain edges to construct weight matrices.')
				
	
	def _check_edge_weight(self, G0, verbose):
		eweight = None
		if self._edge_weight is None:
			if verbose >= 2:
				print('\n None edge weight is specified. Set all weight to 1.\n')
		else:
			try:
				some_weight = list(nx.get_edge_attributes(G0, self._edge_weight).values())[0]
				if isinstance(some_weight, float) or isinstance(some_weight, int):
					eweight = self._edge_weight
				else:
					if verbose >= 2:
						print('\n Edge weight with name %s is not float or integer. Set all weight to 1.\n' % self._edge_weight)
			except:
				if verbose >= 2:
					print('\n Edge weight with name "%s" is not found in the edge attributes. Set all weight to 1.\n' % self._edge_weight)
		
		self._edge_weight = eweight
				
		
	def _add_dummy_labels(self, Gn):
		if len(self._node_labels) == 0 or (len(self._node_labels) == 1 and self._node_labels[0] == SpecialLabel.DUMMY):
			for i in range(len(Gn)):
				nx.set_node_attributes(Gn[i], '0', SpecialLabel.DUMMY)
			self._node_labels = [SpecialLabel.DUMMY]
		if len(self._edge_labels) == 0 or (len(self._edge_labels) == 1 and self._edge_labels[0] == SpecialLabel.DUMMY):
			for i in range(len(Gn)):
				nx.set_edge_attributes(Gn[i], '0', SpecialLabel.DUMMY)
			self._edge_labels = [SpecialLabel.DUMMY]


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Sep 27 10:56:23 2018

@author: linlin

@references:

	[1] Suard F, Rakotomamonjy A, Bensrhair A. Kernel on Bag of Paths For
	Measuring Similarity of Shapes. InESANN 2007 Apr 25 (pp. 355-360).
"""

import sys
import time
from itertools import combinations, product
from functools import partial
from multiprocessing import Pool
from tqdm import tqdm

import networkx as nx
import numpy as np

from gklearn.utils.graphdataset import get_dataset_attributes
from gklearn.utils.parallel import parallel_gm
from gklearn.utils.trie import Trie

def structuralspkernel(*args,
					   node_label='atom',
					   edge_weight=None,
					   edge_label='bond_type',
					   node_kernels=None,
					   edge_kernels=None,
					   compute_method='naive',
					   parallel='imap_unordered',
#					   parallel=None,
					   n_jobs=None,
					   chunksize=None,
					   verbose=True):
	"""Compute mean average structural shortest path kernels between graphs.

	Parameters
	----------
	Gn : List of NetworkX graph
		List of graphs between which the kernels are computed.

	G1, G2 : NetworkX graphs
		Two graphs between which the kernel is computed.

	node_label : string
		Node attribute used as label. The default node label is atom.

	edge_weight : string
		Edge attribute name corresponding to the edge weight. Applied for the
		computation of the shortest paths.

	edge_label : string
		Edge attribute used as label. The default edge label is bond_type.

	node_kernels : dict
		A dictionary of kernel functions for nodes, including 3 items: 'symb'
		for symbolic node labels, 'nsymb' for non-symbolic node labels, 'mix'
		for both labels. The first 2 functions take two node labels as
		parameters, and the 'mix' function takes 4 parameters, a symbolic and a
		non-symbolic label for each the two nodes. Each label is in form of 2-D
		dimension array (n_samples, n_features). Each function returns a number
		as the kernel value. Ignored when nodes are unlabeled.

	edge_kernels : dict
		A dictionary of kernel functions for edges, including 3 items: 'symb'
		for symbolic edge labels, 'nsymb' for non-symbolic edge labels, 'mix'
		for both labels. The first 2 functions take two edge labels as
		parameters, and the 'mix' function takes 4 parameters, a symbolic and a
		non-symbolic label for each the two edges. Each label is in form of 2-D
		dimension array (n_samples, n_features). Each function returns a number
		as the kernel value. Ignored when edges are unlabeled.

	compute_method : string
		Computation method to store the shortest paths and compute the graph
		kernel. The Following choices are available:

		'trie': store paths as tries.

		'naive': store paths to lists.

	n_jobs : int
		Number of jobs for parallelization.

	Return
	------
	Kmatrix : Numpy matrix
		Kernel matrix, each element of which is the mean average structural
		shortest path kernel between 2 praphs.
	"""
	# pre-process
	Gn = args[0] if len(args) == 1 else [args[0], args[1]]
	Gn = [g.copy() for g in Gn]
	weight = None
	if edge_weight is None:
		if verbose:
			print('\n None edge weight specified. Set all weight to 1.\n')
	else:
		try:
			some_weight = list(
				nx.get_edge_attributes(Gn[0], edge_weight).values())[0]
			if isinstance(some_weight, (float, int)):
				weight = edge_weight
			else:
				if verbose:
					print(
							'\n Edge weight with name %s is not float or integer. Set all weight to 1.\n'
							% edge_weight)
		except:
			if verbose:
				print(
						'\n Edge weight with name "%s" is not found in the edge attributes. Set all weight to 1.\n'
						% edge_weight)
	ds_attrs = get_dataset_attributes(
		Gn,
		attr_names=['node_labeled', 'node_attr_dim', 'edge_labeled',
					'edge_attr_dim', 'is_directed'],
		node_label=node_label, edge_label=edge_label)

	start_time = time.time()

	# get shortest paths of each graph in Gn
	if parallel == 'imap_unordered':
		splist = [None] * len(Gn)
		pool = Pool(n_jobs)
		itr = zip(Gn, range(0, len(Gn)))
		if chunksize is None:
			if len(Gn) < 100 * n_jobs:
				chunksize = int(len(Gn) / n_jobs) + 1
			else:
				chunksize = 100
		# get shortest path graphs of Gn
		if compute_method == 'trie':
			getsp_partial = partial(wrapper_getSP_trie, weight, ds_attrs['is_directed'])
		else:
			getsp_partial = partial(wrapper_getSP_naive, weight, ds_attrs['is_directed'])
		if verbose:
			iterator = tqdm(pool.imap_unordered(getsp_partial, itr, chunksize),
							desc='getting shortest paths', file=sys.stdout)
		else:
			iterator = pool.imap_unordered(getsp_partial, itr, chunksize)
		for i, sp in iterator:
			splist[i] = sp
	#		time.sleep(10)
		pool.close()
		pool.join()
	# ---- direct running, normally use single CPU core. ----
	elif parallel is None:
		splist = []
		if verbose:
			iterator = tqdm(Gn, desc='getting sp graphs', file=sys.stdout)
		else:
			iterator = Gn
		if compute_method == 'trie':
			for g in iterator:
				splist.append(get_sps_as_trie(g, weight, ds_attrs['is_directed']))
		else:
			for g in iterator:
				splist.append(get_shortest_paths(g, weight, ds_attrs['is_directed']))

#	ss = 0
#	ss += sys.getsizeof(splist)
#	for spss in splist:
#		ss += sys.getsizeof(spss)
#		for spp in spss:
#			ss += sys.getsizeof(spp)


#	time.sleep(20)



	# # ---- only for the Fast Computation of Shortest Path Kernel (FCSP)
	# sp_ml = [0] * len(Gn)  # shortest path matrices
	# for i in result_sp:
	#	 sp_ml[i[0]] = i[1]
	# edge_x_g = [[] for i in range(len(sp_ml))]
	# edge_y_g = [[] for i in range(len(sp_ml))]
	# edge_w_g = [[] for i in range(len(sp_ml))]
	# for idx, item in enumerate(sp_ml):
	#	 for i1 in range(len(item)):
	#		 for i2 in range(i1 + 1, len(item)):
	#			 if item[i1, i2] != np.inf:
	#				 edge_x_g[idx].append(i1)
	#				 edge_y_g[idx].append(i2)
	#				 edge_w_g[idx].append(item[i1, i2])
	# print(len(edge_x_g[0]))
	# print(len(edge_y_g[0]))
	# print(len(edge_w_g[0]))

	Kmatrix = np.zeros((len(Gn), len(Gn)))

	# ---- use pool.imap_unordered to parallel and track progress. ----
	if parallel == 'imap_unordered':
		def init_worker(spl_toshare, gs_toshare):
			global G_spl, G_gs
			G_spl = spl_toshare
			G_gs = gs_toshare
		if compute_method == 'trie':
			do_partial = partial(wrapper_ssp_do_trie, ds_attrs, node_label, edge_label,
								 node_kernels, edge_kernels)
			parallel_gm(do_partial, Kmatrix, Gn, init_worker=init_worker,
								glbv=(splist, Gn), n_jobs=n_jobs, chunksize=chunksize, verbose=verbose)
		else:
			do_partial = partial(wrapper_ssp_do, ds_attrs, node_label, edge_label,
								 node_kernels, edge_kernels)
			parallel_gm(do_partial, Kmatrix, Gn, init_worker=init_worker,
								glbv=(splist, Gn), n_jobs=n_jobs, chunksize=chunksize, verbose=verbose)
	# ---- direct running, normally use single CPU core. ----
	elif parallel is None:
		from itertools import combinations_with_replacement
		itr = combinations_with_replacement(range(0, len(Gn)), 2)
		if verbose:
			iterator = tqdm(itr, desc='Computing kernels', file=sys.stdout)
		else:
			iterator = itr
		if compute_method == 'trie':
			for i, j in iterator:
				kernel = ssp_do_trie(Gn[i], Gn[j], splist[i], splist[j],
						ds_attrs, node_label, edge_label, node_kernels, edge_kernels)
				Kmatrix[i][j] = kernel
				Kmatrix[j][i] = kernel
		else:
			for i, j in iterator:
				kernel = structuralspkernel_do(Gn[i], Gn[j], splist[i], splist[j],
						ds_attrs, node_label, edge_label, node_kernels, edge_kernels)
		#		if(kernel > 1):
		#			print("error here ")
				Kmatrix[i][j] = kernel
				Kmatrix[j][i] = kernel

#	# ---- use pool.map to parallel. ----
#	pool = Pool(n_jobs)
#	do_partial = partial(wrapper_ssp_do, ds_attrs, node_label, edge_label,
#						 node_kernels, edge_kernels)
#	itr = zip(combinations_with_replacement(Gn, 2),
#			  combinations_with_replacement(splist, 2),
#			  combinations_with_replacement(range(0, len(Gn)), 2))
#	for i, j, kernel in tqdm(
#			pool.map(do_partial, itr), desc='Computing kernels',
#			file=sys.stdout):
#		Kmatrix[i][j] = kernel
#		Kmatrix[j][i] = kernel
#	pool.close()
#	pool.join()

#	# ---- use pool.imap_unordered to parallel and track progress. ----
#	do_partial = partial(wrapper_ssp_do, ds_attrs, node_label, edge_label,
#						 node_kernels, edge_kernels)
#	itr = zip(combinations_with_replacement(Gn, 2),
#			  combinations_with_replacement(splist, 2),
#			  combinations_with_replacement(range(0, len(Gn)), 2))
#	len_itr = int(len(Gn) * (len(Gn) + 1) / 2)
#	if len_itr < 1000 * n_jobs:
#		chunksize = int(len_itr / n_jobs) + 1
#	else:
#		chunksize = 1000
#	from contextlib import closing
#	with closing(Pool(n_jobs)) as pool:
#		for i, j, kernel in tqdm(
#				pool.imap_unordered(do_partial, itr, 1000),
#				desc='Computing kernels',
#				file=sys.stdout):
#			Kmatrix[i][j] = kernel
#			Kmatrix[j][i] = kernel
#	pool.close()
#	pool.join()



	run_time = time.time() - start_time
	if verbose:
		print("\n --- shortest path kernel matrix of size %d built in %s seconds ---"
			  % (len(Gn), run_time))

	return Kmatrix, run_time


def structuralspkernel_do(g1, g2, spl1, spl2, ds_attrs, node_label, edge_label,
						  node_kernels, edge_kernels):

	kernel = 0

	# First, compute shortest path matrices, method borrowed from FCSP.
	vk_dict = getAllNodeKernels(g1, g2, node_kernels, node_label, ds_attrs)
	# Then, compute kernels between all pairs of edges, which is an idea of
	# extension of FCSP. It suits sparse graphs, which is the most case we
	# went though. For dense graphs, this would be slow.
	ek_dict = getAllEdgeKernels(g1, g2, edge_kernels, edge_label, ds_attrs)

	# compute graph kernels
	if vk_dict:
		if ek_dict:
			for p1, p2 in product(spl1, spl2):
				if len(p1) == len(p2):
					kpath = vk_dict[(p1[0], p2[0])]
					if kpath:
						for idx in range(1, len(p1)):
							kpath *= vk_dict[(p1[idx], p2[idx])] * \
								ek_dict[((p1[idx-1], p1[idx]),
										 (p2[idx-1], p2[idx]))]
							if not kpath:
								break
						kernel += kpath  # add up kernels of all paths
		else:
			for p1, p2 in product(spl1, spl2):
				if len(p1) == len(p2):
					kpath = vk_dict[(p1[0], p2[0])]
					if kpath:
						for idx in range(1, len(p1)):
							kpath *= vk_dict[(p1[idx], p2[idx])]
							if not kpath:
								break
						kernel += kpath  # add up kernels of all paths
	else:
		if ek_dict:
			for p1, p2 in product(spl1, spl2):
				if len(p1) == len(p2):
					if len(p1) == 0:
						kernel += 1
					else:
						kpath = 1
						for idx in range(0, len(p1) - 1):
							kpath *= ek_dict[((p1[idx], p1[idx+1]),
											  (p2[idx], p2[idx+1]))]
							if not kpath:
								break
						kernel += kpath  # add up kernels of all paths
		else:
			for p1, p2 in product(spl1, spl2):
				if len(p1) == len(p2):
					kernel += 1
	try:
		kernel = kernel / (len(spl1) * len(spl2))  # Compute mean average
	except ZeroDivisionError:
		print(spl1, spl2)
		print(g1.nodes(data=True))
		print(g1.edges(data=True))
		raise Exception

	# # ---- exact implementation of the Fast Computation of Shortest Path Kernel (FCSP), reference [2], sadly it is slower than the current implementation
	# # compute vertex kernel matrix
	# try:
	#	 vk_mat = np.zeros((nx.number_of_nodes(g1),
	#						nx.number_of_nodes(g2)))
	#	 g1nl = enumerate(g1.nodes(data=True))
	#	 g2nl = enumerate(g2.nodes(data=True))
	#	 for i1, n1 in g1nl:
	#		 for i2, n2 in g2nl:
	#			 vk_mat[i1][i2] = kn(
	#				 n1[1][node_label], n2[1][node_label],
	#				 [n1[1]['attributes']], [n2[1]['attributes']])

	#	 range1 = range(0, len(edge_w_g[i]))
	#	 range2 = range(0, len(edge_w_g[j]))
	#	 for i1 in range1:
	#		 x1 = edge_x_g[i][i1]
	#		 y1 = edge_y_g[i][i1]
	#		 w1 = edge_w_g[i][i1]
	#		 for i2 in range2:
	#			 x2 = edge_x_g[j][i2]
	#			 y2 = edge_y_g[j][i2]
	#			 w2 = edge_w_g[j][i2]
	#			 ke = (w1 == w2)
	#			 if ke > 0:
	#				 kn1 = vk_mat[x1][x2] * vk_mat[y1][y2]
	#				 kn2 = vk_mat[x1][y2] * vk_mat[y1][x2]
	#				 Kmatrix += kn1 + kn2
	return kernel


def wrapper_ssp_do(ds_attrs, node_label, edge_label, node_kernels,
				   edge_kernels, itr):
	i = itr[0]
	j = itr[1]
	return i, j, structuralspkernel_do(G_gs[i], G_gs[j], G_spl[i], G_spl[j],
									   ds_attrs, node_label, edge_label,
									   node_kernels, edge_kernels)


def ssp_do_trie(g1, g2, trie1, trie2, ds_attrs, node_label, edge_label,
						  node_kernels, edge_kernels):

#	# traverse all paths in graph1. Deep-first search is applied.
#	def traverseBothTrie(root, trie2, kernel, pcurrent=[]):
#		for key, node in root['children'].items():
#			pcurrent.append(key)
#			if node['isEndOfWord']:
#	#					print(node['count'])
#				traverseTrie2(trie2.root, pcurrent, kernel,
#							  pcurrent=[])
#			if node['children'] != {}:
#				traverseBothTrie(node, trie2, kernel, pcurrent)
#			else:
#				del pcurrent[-1]
#		if pcurrent != []:
#			del pcurrent[-1]
#
#
#	# traverse all paths in graph2 and find out those that are not in
#	# graph1. Deep-first search is applied.
#	def traverseTrie2(root, p1, kernel, pcurrent=[]):
#		for key, node in root['children'].items():
#			pcurrent.append(key)
#			if node['isEndOfWord']:
#	#					print(node['count'])
#				kernel[0] += computePathKernel(p1, pcurrent, vk_dict, ek_dict)
#			if node['children'] != {}:
#				traverseTrie2(node, p1, kernel, pcurrent)
#			else:
#				del pcurrent[-1]
#		if pcurrent != []:
#			del pcurrent[-1]
#
#
#	kernel = [0]
#
#	# First, compute shortest path matrices, method borrowed from FCSP.
#	vk_dict = getAllNodeKernels(g1, g2, node_kernels, node_label, ds_attrs)
#	# Then, compute kernels between all pairs of edges, which is an idea of
#	# extension of FCSP. It suits sparse graphs, which is the most case we
#	# went though. For dense graphs, this would be slow.
#	ek_dict = getAllEdgeKernels(g1, g2, edge_kernels, edge_label, ds_attrs)
#
#	# compute graph kernels
#	traverseBothTrie(trie1[0].root, trie2[0], kernel)
#
#	kernel = kernel[0] / (trie1[1] * trie2[1])  # Compute mean average

#	# traverse all paths in graph1. Deep-first search is applied.
#	def traverseBothTrie(root, trie2, kernel, vk_dict, ek_dict, pcurrent=[]):
#		for key, node in root['children'].items():
#			pcurrent.append(key)
#			if node['isEndOfWord']:
#	#					print(node['count'])
#				traverseTrie2(trie2.root, pcurrent, kernel, vk_dict, ek_dict,
#							  pcurrent=[])
#			if node['children'] != {}:
#				traverseBothTrie(node, trie2, kernel, vk_dict, ek_dict, pcurrent)
#			else:
#				del pcurrent[-1]
#		if pcurrent != []:
#			del pcurrent[-1]
#
#
#	# traverse all paths in graph2 and find out those that are not in
#	# graph1. Deep-first search is applied.
#	def traverseTrie2(root, p1, kernel, vk_dict, ek_dict, pcurrent=[]):
#		for key, node in root['children'].items():
#			pcurrent.append(key)
#			if node['isEndOfWord']:
#	#					print(node['count'])
#				kernel[0] += computePathKernel(p1, pcurrent, vk_dict, ek_dict)
#			if node['children'] != {}:
#				traverseTrie2(node, p1, kernel, vk_dict, ek_dict, pcurrent)
#			else:
#				del pcurrent[-1]
#		if pcurrent != []:
#			del pcurrent[-1]


	kernel = [0]

	# First, compute shortest path matrices, method borrowed from FCSP.
	vk_dict = getAllNodeKernels(g1, g2, node_kernels, node_label, ds_attrs)
	# Then, compute kernels between all pairs of edges, which is an idea of
	# extension of FCSP. It suits sparse graphs, which is the most case we
	# went though. For dense graphs, this would be slow.
	ek_dict = getAllEdgeKernels(g1, g2, edge_kernels, edge_label, ds_attrs)

	# compute graph kernels
#	traverseBothTrie(trie1[0].root, trie2[0], kernel, vk_dict, ek_dict)
	if vk_dict:
		if ek_dict:
			traverseBothTriem(trie1[0].root, trie2[0], kernel, vk_dict, ek_dict)
		else:
			traverseBothTriev(trie1[0].root, trie2[0], kernel, vk_dict, ek_dict)
	else:
		if ek_dict:
			traverseBothTriee(trie1[0].root, trie2[0], kernel, vk_dict, ek_dict)
		else:
			traverseBothTrieu(trie1[0].root, trie2[0], kernel, vk_dict, ek_dict)

	kernel = kernel[0] / (trie1[1] * trie2[1])  # Compute mean average

	return kernel


def wrapper_ssp_do_trie(ds_attrs, node_label, edge_label, node_kernels,
				   edge_kernels, itr):
	i = itr[0]
	j = itr[1]
	return i, j, ssp_do_trie(G_gs[i], G_gs[j], G_spl[i], G_spl[j], ds_attrs,
							 node_label, edge_label, node_kernels, edge_kernels)


def getAllNodeKernels(g1, g2, node_kernels, node_label, ds_attrs):
	# compute shortest path matrices, method borrowed from FCSP.
	vk_dict = {}  # shortest path matrices dict
	if ds_attrs['node_labeled']:
		# node symb and non-synb labeled
		if ds_attrs['node_attr_dim'] > 0:
			kn = node_kernels['mix']
			for n1, n2 in product(
					g1.nodes(data=True), g2.nodes(data=True)):
				vk_dict[(n1[0], n2[0])] = kn(
					n1[1][node_label], n2[1][node_label],
					n1[1]['attributes'], n2[1]['attributes'])
		# node symb labeled
		else:
			kn = node_kernels['symb']
			for n1 in g1.nodes(data=True):
				for n2 in g2.nodes(data=True):
					vk_dict[(n1[0], n2[0])] = kn(n1[1][node_label],
												 n2[1][node_label])
	else:
		# node non-synb labeled
		if ds_attrs['node_attr_dim'] > 0:
			kn = node_kernels['nsymb']
			for n1 in g1.nodes(data=True):
				for n2 in g2.nodes(data=True):
					vk_dict[(n1[0], n2[0])] = kn(n1[1]['attributes'],
												 n2[1]['attributes'])
		# node unlabeled
		else:
			pass

	return vk_dict


def getAllEdgeKernels(g1, g2, edge_kernels, edge_label, ds_attrs):
	# compute kernels between all pairs of edges, which is an idea of
	# extension of FCSP. It suits sparse graphs, which is the most case we
	# went though. For dense graphs, this would be slow.
	ek_dict = {}  # dict of edge kernels
	if ds_attrs['edge_labeled']:
		# edge symb and non-synb labeled
		if ds_attrs['edge_attr_dim'] > 0:
			ke = edge_kernels['mix']
			for e1, e2 in product(
					g1.edges(data=True), g2.edges(data=True)):
				ek_temp = ke(e1[2][edge_label], e2[2][edge_label],
					e1[2]['attributes'], e2[2]['attributes'])
				ek_dict[((e1[0], e1[1]), (e2[0], e2[1]))] = ek_temp
				ek_dict[((e1[1], e1[0]), (e2[0], e2[1]))] = ek_temp
				ek_dict[((e1[0], e1[1]), (e2[1], e2[0]))] = ek_temp
				ek_dict[((e1[1], e1[0]), (e2[1], e2[0]))] = ek_temp
		# edge symb labeled
		else:
			ke = edge_kernels['symb']
			for e1 in g1.edges(data=True):
				for e2 in g2.edges(data=True):
					ek_temp = ke(e1[2][edge_label], e2[2][edge_label])
					ek_dict[((e1[0], e1[1]), (e2[0], e2[1]))] = ek_temp
					ek_dict[((e1[1], e1[0]), (e2[0], e2[1]))] = ek_temp
					ek_dict[((e1[0], e1[1]), (e2[1], e2[0]))] = ek_temp
					ek_dict[((e1[1], e1[0]), (e2[1], e2[0]))] = ek_temp
	else:
		# edge non-synb labeled
		if ds_attrs['edge_attr_dim'] > 0:
			ke = edge_kernels['nsymb']
			for e1 in g1.edges(data=True):
				for e2 in g2.edges(data=True):
					ek_temp = ke(e1[2]['attributes'], e2[2]['attributes'])
					ek_dict[((e1[0], e1[1]), (e2[0], e2[1]))] = ek_temp
					ek_dict[((e1[1], e1[0]), (e2[0], e2[1]))] = ek_temp
					ek_dict[((e1[0], e1[1]), (e2[1], e2[0]))] = ek_temp
					ek_dict[((e1[1], e1[0]), (e2[1], e2[0]))] = ek_temp
		# edge unlabeled
		else:
			pass

	return ek_dict


# traverse all paths in graph1. Deep-first search is applied.
def traverseBothTriem(root, trie2, kernel, vk_dict, ek_dict, pcurrent=[]):
	for key, node in root['children'].items():
		pcurrent.append(key)
		if node['isEndOfWord']:
#					print(node['count'])
			traverseTrie2m(trie2.root, pcurrent, kernel, vk_dict, ek_dict,
						  pcurrent=[])
		if node['children'] != {}:
			traverseBothTriem(node, trie2, kernel, vk_dict, ek_dict, pcurrent)
		else:
			del pcurrent[-1]
	if pcurrent != []:
		del pcurrent[-1]


# traverse all paths in graph2 and find out those that are not in
# graph1. Deep-first search is applied.
def traverseTrie2m(root, p1, kernel, vk_dict, ek_dict, pcurrent=[]):
	for key, node in root['children'].items():
		pcurrent.append(key)
		if node['isEndOfWord']:
#					print(node['count'])
			if len(p1) == len(pcurrent):
				kpath = vk_dict[(p1[0], pcurrent[0])]
				if kpath:
					for idx in range(1, len(p1)):
						kpath *= vk_dict[(p1[idx], pcurrent[idx])] * \
							ek_dict[((p1[idx-1], p1[idx]),
									 (pcurrent[idx-1], pcurrent[idx]))]
						if not kpath:
							break
					kernel[0] += kpath  # add up kernels of all paths
		if node['children'] != {}:
			traverseTrie2m(node, p1, kernel, vk_dict, ek_dict, pcurrent)
		else:
			del pcurrent[-1]
	if pcurrent != []:
		del pcurrent[-1]


# traverse all paths in graph1. Deep-first search is applied.
def traverseBothTriev(root, trie2, kernel, vk_dict, ek_dict, pcurrent=[]):
	for key, node in root['children'].items():
		pcurrent.append(key)
		if node['isEndOfWord']:
#					print(node['count'])
			traverseTrie2v(trie2.root, pcurrent, kernel, vk_dict, ek_dict,
						  pcurrent=[])
		if node['children'] != {}:
			traverseBothTriev(node, trie2, kernel, vk_dict, ek_dict, pcurrent)
		else:
			del pcurrent[-1]
	if pcurrent != []:
		del pcurrent[-1]


# traverse all paths in graph2 and find out those that are not in
# graph1. Deep-first search is applied.
def traverseTrie2v(root, p1, kernel, vk_dict, ek_dict, pcurrent=[]):
	for key, node in root['children'].items():
		pcurrent.append(key)
		if node['isEndOfWord']:
#					print(node['count'])
			if len(p1) == len(pcurrent):
				kpath = vk_dict[(p1[0], pcurrent[0])]
				if kpath:
					for idx in range(1, len(p1)):
						kpath *= vk_dict[(p1[idx], pcurrent[idx])]
						if not kpath:
							break
					kernel[0] += kpath  # add up kernels of all paths
		if node['children'] != {}:
			traverseTrie2v(node, p1, kernel, vk_dict, ek_dict, pcurrent)
		else:
			del pcurrent[-1]
	if pcurrent != []:
		del pcurrent[-1]


# traverse all paths in graph1. Deep-first search is applied.
def traverseBothTriee(root, trie2, kernel, vk_dict, ek_dict, pcurrent=[]):
	for key, node in root['children'].items():
		pcurrent.append(key)
		if node['isEndOfWord']:
#					print(node['count'])
			traverseTrie2e(trie2.root, pcurrent, kernel, vk_dict, ek_dict,
						  pcurrent=[])
		if node['children'] != {}:
			traverseBothTriee(node, trie2, kernel, vk_dict, ek_dict, pcurrent)
		else:
			del pcurrent[-1]
	if pcurrent != []:
		del pcurrent[-1]


# traverse all paths in graph2 and find out those that are not in
# graph1. Deep-first search is applied.
def traverseTrie2e(root, p1, kernel, vk_dict, ek_dict, pcurrent=[]):
	for key, node in root['children'].items():
		pcurrent.append(key)
		if node['isEndOfWord']:
#					print(node['count'])
			if len(p1) == len(pcurrent):
				if len(p1) == 0:
					kernel += 1
				else:
					kpath = 1
					for idx in range(0, len(p1) - 1):
						kpath *= ek_dict[((p1[idx], p1[idx+1]),
										  (pcurrent[idx], pcurrent[idx+1]))]
						if not kpath:
							break
					kernel[0] += kpath  # add up kernels of all paths
		if node['children'] != {}:
			traverseTrie2e(node, p1, kernel, vk_dict, ek_dict, pcurrent)
		else:
			del pcurrent[-1]
	if pcurrent != []:
		del pcurrent[-1]


# traverse all paths in graph1. Deep-first search is applied.
def traverseBothTrieu(root, trie2, kernel, vk_dict, ek_dict, pcurrent=[]):
	for key, node in root['children'].items():
		pcurrent.append(key)
		if node['isEndOfWord']:
#					print(node['count'])
			traverseTrie2u(trie2.root, pcurrent, kernel, vk_dict, ek_dict,
						  pcurrent=[])
		if node['children'] != {}:
			traverseBothTrieu(node, trie2, kernel, vk_dict, ek_dict, pcurrent)
		else:
			del pcurrent[-1]
	if pcurrent != []:
		del pcurrent[-1]


# traverse all paths in graph2 and find out those that are not in
# graph1. Deep-first search is applied.
def traverseTrie2u(root, p1, kernel, vk_dict, ek_dict, pcurrent=[]):
	for key, node in root['children'].items():
		pcurrent.append(key)
		if node['isEndOfWord']:
#					print(node['count'])
			if len(p1) == len(pcurrent):
				kernel[0] += 1
		if node['children'] != {}:
			traverseTrie2u(node, p1, kernel, vk_dict, ek_dict, pcurrent)
		else:
			del pcurrent[-1]
	if pcurrent != []:
		del pcurrent[-1]


#def computePathKernel(p1, p2, vk_dict, ek_dict):
#	kernel = 0
#	if vk_dict:
#		if ek_dict:
#			if len(p1) == len(p2):
#				kpath = vk_dict[(p1[0], p2[0])]
#				if kpath:
#					for idx in range(1, len(p1)):
#						kpath *= vk_dict[(p1[idx], p2[idx])] * \
#							ek_dict[((p1[idx-1], p1[idx]),
#									 (p2[idx-1], p2[idx]))]
#						if not kpath:
#							break
#					kernel += kpath  # add up kernels of all paths
#		else:
#			if len(p1) == len(p2):
#				kpath = vk_dict[(p1[0], p2[0])]
#				if kpath:
#					for idx in range(1, len(p1)):
#						kpath *= vk_dict[(p1[idx], p2[idx])]
#						if not kpath:
#							break
#					kernel += kpath  # add up kernels of all paths
#	else:
#		if ek_dict:
#			if len(p1) == len(p2):
#				if len(p1) == 0:
#					kernel += 1
#				else:
#					kpath = 1
#					for idx in range(0, len(p1) - 1):
#						kpath *= ek_dict[((p1[idx], p1[idx+1]),
#										  (p2[idx], p2[idx+1]))]
#						if not kpath:
#							break
#					kernel += kpath  # add up kernels of all paths
#		else:
#			if len(p1) == len(p2):
#				kernel += 1
#
#	return kernel


def get_shortest_paths(G, weight, directed):
	"""Get all shortest paths of a graph.

	Parameters
	----------
	G : NetworkX graphs
		The graphs whose paths are computed.
	weight : string/None
		edge attribute used as weight to compute the shortest path.
	directed: boolean
		Whether graph is directed.

	Return
	------
	sp : list of list
		List of shortest paths of the graph, where each path is represented by a list of nodes.
	"""
	sp = []
	for n1, n2 in combinations(G.nodes(), 2):
		try:
			spltemp = list(nx.all_shortest_paths(G, n1, n2, weight=weight))
		except nx.NetworkXNoPath:  # nodes not connected
			#			sp.append([])
			pass
		else:
			sp += spltemp
			# each edge walk is counted twice, starting from both its extreme nodes.
			if not directed:
				sp += [sptemp[::-1] for sptemp in spltemp]

	# add single nodes as length 0 paths.
	sp += [[n] for n in G.nodes()]
	return sp


def wrapper_getSP_naive(weight, directed, itr_item):
	g = itr_item[0]
	i = itr_item[1]
	return i, get_shortest_paths(g, weight, directed)


def get_sps_as_trie(G, weight, directed):
	"""Get all shortest paths of a graph and insert them into a trie.

	Parameters
	----------
	G : NetworkX graphs
		The graphs whose paths are computed.
	weight : string/None
		edge attribute used as weight to compute the shortest path.
	directed: boolean
		Whether graph is directed.

	Return
	------
	sp : list of list
		List of shortest paths of the graph, where each path is represented by a list of nodes.
	"""
	sptrie = Trie()
	lensp = 0
	for n1, n2 in combinations(G.nodes(), 2):
		try:
			spltemp = list(nx.all_shortest_paths(G, n1, n2, weight=weight))
		except nx.NetworkXNoPath:  # nodes not connected
			pass
		else:
			lensp += len(spltemp)
			if not directed:
				lensp += len(spltemp)
			for sp in spltemp:
				sptrie.insertWord(sp)
			# each edge walk is counted twice, starting from both its extreme nodes.
				if not directed:
					sptrie.insertWord(sp[::-1])

	# add single nodes as length 0 paths.
	for n in G.nodes():
		sptrie.insertWord([n])

	return sptrie, lensp + nx.number_of_nodes(G)


def wrapper_getSP_trie(weight, directed, itr_item):
	g = itr_item[0]
	i = itr_item[1]
	return i, get_sps_as_trie(g, weight, directed)
